"""This module contains the dynamic loader logic to get on-chain storage data
and dependencies."""
from mythril.disassembler.disassembly import Disassembly
import logging
import re
import functools
from mythril.ethereum.interface.rpc.client import EthJsonRpc
from typing import Optional

LRU_CACHE_SIZE = 4096

log = logging.getLogger(__name__)


class DynLoader:
    """The dynamic loader class."""

    def __init__(self, eth: Optional[EthJsonRpc], active=True):
        """

        :param eth:
        :param active:
        """
        self.eth = eth
        self.active = active

    @functools.lru_cache(LRU_CACHE_SIZE)
    def read_storage(self, contract_address: str, index: int) -> str:
        """

        :param contract_address:
        :param index:
        :return:
        """
        if not self.active:
            raise ValueError("Loader is disabled")
        if not self.eth:
            raise ValueError("Cannot load from the storage when eth is None")

        value = self.eth.eth_getStorageAt(
            contract_address, position=index, block="latest"
        )
        if value == "0x":
            value = "0x0000000000000000000000000000000000000000000000000000000000000000"
        return value

    @functools.lru_cache(LRU_CACHE_SIZE)
    def read_balance(self, address: str) -> str:
        """

        :param address:
        :return:
        """
        if not self.active:
            raise ValueError("Cannot load from storage when the loader is disabled")
        if not self.eth:
            raise ValueError(
                "Cannot load from the chain when eth is None, please use rpc, or specify infura-id"
            )

        return self.eth.eth_getBalance(address)

    @functools.lru_cache(LRU_CACHE_SIZE)
    def dynld(self, dependency_address: str) -> Optional[Disassembly]:
        """
        :param dependency_address:
        :return:
        """
        if not self.active:
            raise ValueError("Loader is disabled")
        if not self.eth:
            raise ValueError(
                "Cannot load from the chain when eth is None, please use rpc, or specify infura-id"
            )

        log.debug("Dynld at contract %s", dependency_address)

        # Ensure that dependency_address is the correct length, with 0s prepended as needed.

        if isinstance(dependency_address, int):
            dependency_address = "0x{:040X}".format(dependency_address)
        else:
            dependency_address = (
                "0x" + "0" * (42 - len(dependency_address)) + dependency_address[2:]
            )

        m = re.match(r"^(0x[0-9a-fA-F]{40})$", dependency_address)

        if m:
            dependency_address = m.group(1)

        else:
            return None

        log.debug("Dependency address: %s", dependency_address)

        code = self.eth.eth_getCode(dependency_address)

        if code == "0x":
            return None
        else:
            return Disassembly(code)



"""This module contains the detection code for predictable variable
dependence."""
import logging

from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.analysis.report import Issue
from mythril.exceptions import UnsatError
from mythril.analysis import solver
from mythril.laser.smt import And, ULT, symbol_factory
from mythril.analysis.swc_data import TIMESTAMP_DEPENDENCE, WEAK_RANDOMNESS
from mythril.analysis.module.module_helpers import is_prehook
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.annotation import StateAnnotation
from typing import cast, List

log = logging.getLogger(__name__)

predictable_ops = ["COINBASE", "GASLIMIT", "TIMESTAMP", "NUMBER"]


class PredictableValueAnnotation:
    """Symbol annotation used if a variable is initialized from a predictable environment variable."""

    def __init__(self, operation: str) -> None:
        self.operation = operation


class OldBlockNumberUsedAnnotation(StateAnnotation):
    """Symbol annotation used if a variable is initialized from a predictable environment variable."""

    def __init__(self) -> None:
        pass


class PredictableVariables(DetectionModule):
    """This module detects whether control flow decisions are made using predictable
    parameters."""

    name = "Control flow depends on a predictable environment variable"
    swc_id = "{} {}".format(TIMESTAMP_DEPENDENCE, WEAK_RANDOMNESS)
    description = (
        "Check whether control flow decisions are influenced by block.coinbase,"
        "block.gaslimit, block.timestamp or block.number."
    )
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["JUMPI", "BLOCKHASH"]
    post_hooks = ["BLOCKHASH"] + predictable_ops

    def _execute(self, state: GlobalState) -> List[Issue]:
        """

        :param state:
        :return:
        """
        return self._analyze_state(state)

    def _analyze_state(self, state: GlobalState) -> List[Issue]:
        """

        :param state:
        :return:
        """

        issues = []

        if is_prehook():

            opcode = state.get_current_instruction()["opcode"]

            if opcode == "JUMPI":

                # Look for predictable state variables in jump condition

                for annotation in state.mstate.stack[-2].annotations:

                    if isinstance(annotation, PredictableValueAnnotation):

                        constraints = state.world_state.constraints
                        try:
                            transaction_sequence = solver.get_transaction_sequence(
                                state, constraints
                            )
                        except UnsatError:
                            continue
                        description = (
                            annotation.operation
                            + " is used to determine a control flow decision. "
                        )
                        description += (
                            "Note that the values of variables like coinbase, gaslimit, block number and timestamp are "
                            "predictable and can be manipulated by a malicious miner. Also keep in mind that "
                            "attackers know hashes of earlier blocks. Don't use any of those environment variables "
                            "as sources of randomness and be aware that use of these variables introduces "
                            "a certain level of trust into miners."
                        )

                        """
                        Usually report low severity except in cases where the hash of a previous block is used to
                        determine control flow. 
                        """

                        severity = "Low"

                        swc_id = (
                            TIMESTAMP_DEPENDENCE
                            if "timestamp" in annotation.operation
                            else WEAK_RANDOMNESS
                        )

                        issue = Issue(
                            contract=state.environment.active_account.contract_name,
                            function_name=state.environment.active_function_name,
                            address=state.get_current_instruction()["address"],
                            swc_id=swc_id,
                            bytecode=state.environment.code.bytecode,
                            title="Dependence on predictable environment variable",
                            severity=severity,
                            description_head="A control flow decision is made based on {}.".format(
                                annotation.operation
                            ),
                            description_tail=description,
                            gas_used=(
                                state.mstate.min_gas_used,
                                state.mstate.max_gas_used,
                            ),
                            transaction_sequence=transaction_sequence,
                        )
                        state.annotate(
                            IssueAnnotation(
                                conditions=[And(*constraints)],
                                issue=issue,
                                detector=self,
                            )
                        )
                        issues.append(issue)

            elif opcode == "BLOCKHASH":

                param = state.mstate.stack[-1]

                constraint = [
                    ULT(param, state.environment.block_number),
                    ULT(
                        state.environment.block_number,
                        symbol_factory.BitVecVal(2**255, 256),
                    ),
                ]

                # Why the second constraint? Because without it Z3 returns a solution where param overflows.

                try:

                    solver.get_model(
                        state.world_state.constraints + constraint  # type: ignore
                    )

                    state.annotate(OldBlockNumberUsedAnnotation())

                except UnsatError:
                    pass

        else:
            # we're in post hook

            opcode = state.environment.code.instruction_list[state.mstate.pc - 1][
                "opcode"
            ]

            if opcode == "BLOCKHASH":
                # if we're in the post hook of a BLOCKHASH op, check if an old block number was used to create it.

                annotations = cast(
                    List[OldBlockNumberUsedAnnotation],
                    list(state.get_annotations(OldBlockNumberUsedAnnotation)),
                )

                if len(annotations):
                    # We can append any block constraint here
                    state.mstate.stack[-1].annotate(
                        PredictableValueAnnotation("The block hash of a previous block")
                    )
            else:
                # Always create an annotation when COINBASE, GASLIMIT, TIMESTAMP or NUMBER is executed.

                state.mstate.stack[-1].annotate(
                    PredictableValueAnnotation(
                        "The block.{} environment variable".format(opcode.lower())
                    )
                )

        return issues


detector = PredictableVariables()



from typing import List
from mythril.support.support_utils import Singleton


class Args(object, metaclass=Singleton):
    """
    This module helps in preventing args being sent through multiple of classes to reach
    any analysis/laser module
    """

    def __init__(self):
        self.solver_timeout = 10000
        self.pruning_factor = None
        self.unconstrained_storage = False
        self.parallel_solving = False
        self.call_depth_limit = 3
        self.disable_iprof = False
        self.solver_log = None
        self.transaction_sequences: List[List[str]] = None
        self.use_integer_module = True
        self.use_issue_annotations = False
        self.solc_args = None
        self.disable_coverage_strategy = False
        self.disable_mutation_pruner = False


args = Args()



"""This module provides classes that make up an issue report."""
import logging
import re
import json
import operator
from eth_abi import decode_abi
from jinja2 import PackageLoader, Environment
from typing import Dict, List, Any, Optional
import hashlib

from mythril.laser.execution_info import ExecutionInfo
from mythril.solidity.soliditycontract import SolidityContract
from mythril.analysis.swc_data import SWC_TO_TITLE
from mythril.support.source_support import Source
from mythril.support.start_time import StartTime
from mythril.support.support_utils import get_code_hash
from mythril.support.signatures import SignatureDB
from time import time

log = logging.getLogger(__name__)


class Issue:
    """Representation of an issue and its location."""

    def __init__(
        self,
        contract: str,
        function_name: str,
        address: int,
        swc_id: str,
        title: str,
        bytecode: str,
        gas_used=(None, None),
        severity=None,
        description_head="",
        description_tail="",
        transaction_sequence=None,
        source_location=None,
    ):
        """

        :param contract: The contract
        :param function_name: Function name where the issue is detected
        :param address: The address of the issue
        :param swc_id: Issue's corresponding swc-id
        :param title: Title
        :param bytecode: bytecode of the issue
        :param gas_used: amount of gas used
        :param severity: The severity of the issue
        :param description_head: The top part of description
        :param description_tail: The bottom part of the description
        :param transaction_sequence: The transaction sequence
        """
        self.title = title
        self.contract = contract
        self.function = function_name
        self.address = address
        self.description_head = description_head
        self.description_tail = description_tail
        self.description = "%s\n%s" % (description_head, description_tail)
        self.severity = severity
        self.swc_id = swc_id
        self.min_gas_used, self.max_gas_used = gas_used
        self.filename = None
        self.code = None
        self.lineno = None
        self.source_mapping = None
        self.discovery_time = time() - StartTime().global_start_time
        self.bytecode_hash = get_code_hash(bytecode)
        self.transaction_sequence = transaction_sequence
        self.source_location = source_location

    @property
    def transaction_sequence_users(self):
        """Returns the transaction sequence without pre-generated block data"""
        return self.transaction_sequence

    @property
    def transaction_sequence_jsonv2(self):
        """Returns the transaction sequence as a json string with pre-generated block data"""
        return (
            self.add_block_data(self.transaction_sequence)
            if self.transaction_sequence
            else None
        )

    @staticmethod
    def add_block_data(transaction_sequence: Dict):
        """Adds sane block data to a transaction_sequence"""
        for step in transaction_sequence["steps"]:
            step["gasLimit"] = "0x7d000"
            step["gasPrice"] = "0x773594000"
            step["blockCoinbase"] = "0xcbcbcbcbcbcbcbcbcbcbcbcbcbcbcbcbcbcbcbcb"
            step["blockDifficulty"] = "0xa7d7343662e26"
            step["blockGasLimit"] = "0x7d0000"
            step["blockNumber"] = "0x66e393"
            step["blockTime"] = "0x5bfa4639"
        return transaction_sequence

    @property
    def as_dict(self):
        """

        :return:
        """

        issue = {
            "title": self.title,
            "swc-id": self.swc_id,
            "contract": self.contract,
            "description": self.description,
            "function": self.function,
            "severity": self.severity,
            "address": self.address,
            "tx_sequence": self.transaction_sequence,
            "min_gas_used": self.min_gas_used,
            "max_gas_used": self.max_gas_used,
            "sourceMap": self.source_mapping,
        }

        if self.filename and self.lineno:
            issue["filename"] = self.filename
            issue["lineno"] = self.lineno

        if self.code:
            issue["code"] = self.code

        return issue

    def _set_internal_compiler_error(self):
        """
        Adds the false positive to description and changes severity to low
        """
        self.severity = "Low"
        self.description_tail += (
            " This issue is reported for internal compiler generated code."
        )
        self.description = "%s\n%s" % (self.description_head, self.description_tail)
        self.code = ""

    def add_code_info(self, contract):
        """

        :param contract:
        """
        if self.address and isinstance(contract, SolidityContract):
            is_constructor = False
            if self.function == "constructor":
                is_constructor = True

            if self.source_location:
                codeinfo = contract.get_source_info(
                    self.source_location, constructor=is_constructor
                )
            else:
                codeinfo = contract.get_source_info(
                    self.address, constructor=is_constructor
                )

            if codeinfo is None:
                self.source_mapping = self.address
                self.filename = "Internal File"
                return

            self.filename = codeinfo.filename
            self.code = codeinfo.code
            self.lineno = codeinfo.lineno
            if self.lineno is None:
                self._set_internal_compiler_error()
            self.source_mapping = codeinfo.solc_mapping
        else:
            self.source_mapping = self.address

    @staticmethod
    def decode_bytes(val):
        if isinstance(val, bytes):
            return val.decode()
        elif isinstance(val, list) or isinstance(val, tuple):
            return [Issue.decode_bytes(x) for x in val]
        else:
            return val

    def resolve_function_names(self):
        """Resolves function names for each step"""

        if (
            self.transaction_sequence is None
            or "steps" not in self.transaction_sequence
        ):
            return

        signatures = SignatureDB()

        for step in self.transaction_sequence["steps"]:
            _hash = step["input"][:10]

            try:
                sig = signatures.get(_hash)
                # TODO: Check other mythx tools for dependency before supporting multiple possible function names
                if len(sig) > 0:
                    step["name"] = sig[0]
                    step["resolved_input"] = Issue.resolve_input(
                        step["calldata"], sig[0]
                    )
                    if step["resolved_input"] is not None:
                        step["resolved_input"] = list(step["resolved_input"])
                        for i, val in enumerate(step["resolved_input"]):
                            step["resolved_input"][i] = Issue.decode_bytes(val)

                        step["resolved_input"] = tuple(step["resolved_input"])

                else:
                    step["name"] = "unknown"
            except ValueError:
                step["name"] = "unknown"

    @staticmethod
    def resolve_input(data, function_name):
        """
        Adds decoded calldate to the tx sequence.
        """
        data = data[10:]

        # Eliminates the first and last brackets
        # Since signature such as func((bytes32,bytes32,uint8)[],(address[],uint32)) are valid
        type_info = function_name[function_name.find("(") + 1 : -1]
        type_info = re.split(r",\s*(?![^()]*\))", type_info)

        if len(data) % 64 > 0:
            data += "0" * (64 - len(data) % 64)
        try:
            decoded_output = decode_abi(type_info, bytes.fromhex(data))
            return decoded_output
        except Exception as e:
            return None


class Report:
    """A report containing the content of multiple issues."""

    environment = Environment(
        loader=PackageLoader("mythril.analysis"), trim_blocks=True
    )

    def __init__(
        self,
        contracts=None,
        exceptions=None,
        execution_info: Optional[List[ExecutionInfo]] = None,
    ):
        """

        :param contracts:
        :param exceptions:
        """
        self.issues = {}  # type: Dict[bytes, Issue]
        self.solc_version = ""
        self.meta = {}  # type: Dict[str, Any]
        self.source = Source()
        self.source.get_source_from_contracts_list(contracts)
        self.exceptions = exceptions or []
        self.execution_info = execution_info or []

    def sorted_issues(self):
        """

        :return:
        """
        issue_list = [issue.as_dict for key, issue in self.issues.items()]
        return sorted(issue_list, key=operator.itemgetter("address", "title"))

    def append_issue(self, issue):
        """

        :param issue:
        """
        m = hashlib.md5()
        m.update(
            (issue.contract + issue.function + str(issue.address) + issue.title).encode(
                "utf-8"
            )
        )
        issue.resolve_function_names()
        self.issues[m.digest()] = issue

    def as_text(self):
        """

        :return:
        """
        name = self._file_name()
        template = Report.environment.get_template("report_as_text.jinja2")

        return template.render(filename=name, issues=self.sorted_issues())

    def as_json(self):
        """

        :return:
        """
        result = {"success": True, "error": None, "issues": self.sorted_issues()}

        return json.dumps(result, sort_keys=True)

    def _get_exception_data(self) -> dict:
        if not self.exceptions:
            return {}
        logs = []  # type: List[Dict]
        for exception in self.exceptions:
            logs += [{"level": "error", "hidden": True, "msg": exception}]
        return {"logs": logs}

    def as_swc_standard_format(self):
        """Format defined for integration and correlation.

        :return:
        """
        # Setup issues
        _issues = []

        for _, issue in self.issues.items():

            idx = self.source.get_source_index(issue.bytecode_hash)
            try:
                title = SWC_TO_TITLE[issue.swc_id]
            except KeyError:
                title = "Unspecified Security Issue"
            extra = {"discoveryTime": int(issue.discovery_time * 10**9)}
            if issue.transaction_sequence_jsonv2:
                extra["testCases"] = [issue.transaction_sequence_jsonv2]

            _issues.append(
                {
                    "swcID": "SWC-" + issue.swc_id,
                    "swcTitle": title,
                    "description": {
                        "head": issue.description_head,
                        "tail": issue.description_tail,
                    },
                    "severity": issue.severity,
                    "locations": [{"sourceMap": "%d:1:%d" % (issue.address, idx)}],
                    "extra": extra,
                }
            )
        # Setup meta
        meta_data = self.meta

        # Add logs to meta
        meta_data.update(self._get_exception_data())

        # Add execution info to meta
        analysis_duration = int(
            round((time() - StartTime().global_start_time) * (10**9))
        )
        meta_data["mythril_execution_info"] = {"analysis_duration": analysis_duration}
        for execution_info in self.execution_info:
            meta_data["mythril_execution_info"].update(execution_info.as_dict())

        result = [
            {
                "issues": _issues,
                "sourceType": self.source.source_type,
                "sourceFormat": self.source.source_format,
                "sourceList": self.source.source_list,
                "meta": meta_data,
            }
        ]

        return json.dumps(result, sort_keys=True)

    def as_markdown(self):
        """

        :return:
        """
        filename = self._file_name()
        template = Report.environment.get_template("report_as_markdown.jinja2")
        return template.render(filename=filename, issues=self.sorted_issues())

    def _file_name(self):
        """

        :return:
        """
        if len(self.issues.values()) > 0:
            return list(self.issues.values())[0].filename



import json
import binascii

from datetime import datetime, timedelta
from typing import Dict, List, Any
from copy import deepcopy

from mythril.concolic.concrete_data import ConcreteData
from mythril.concolic.find_trace import concrete_execution
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.strategy.concolic import ConcolicStrategy
from mythril.laser.ethereum.svm import LaserEVM
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.transaction.symbolic import execute_transaction
from mythril.laser.ethereum.transaction.transaction_models import tx_id_manager
from mythril.laser.smt import Expression, BitVec, symbol_factory
from mythril.laser.ethereum.time_handler import time_handler
from mythril.support.support_args import args


def flip_branches(
    init_state: WorldState,
    concrete_data: ConcreteData,
    jump_addresses: List[str],
    trace: List,
) -> List[Dict[str, Dict[str, Any]]]:
    """
    Flips branches and prints the input required for branch flip

    :param concrete_data: Concrete data
    :param jump_addresses: Jump addresses to flip
    :param trace: trace to follow
    """
    tx_id_manager.restart_counter()
    output_list = []
    laser_evm = LaserEVM(
        execution_timeout=600, use_reachability_check=False, transaction_count=10
    )
    laser_evm.open_states = [deepcopy(init_state)]
    laser_evm.strategy = ConcolicStrategy(
        work_list=laser_evm.work_list,
        max_depth=100,
        trace=trace,
        flip_branch_addresses=jump_addresses,
    )

    time_handler.start_execution(laser_evm.execution_timeout)
    laser_evm.time = datetime.now()

    for transaction in concrete_data["steps"]:
        execute_transaction(
            laser_evm,
            callee_address=transaction["address"],
            caller_address=symbol_factory.BitVecVal(
                int(transaction["origin"], 16), 256
            ),
            data=transaction["input"][2:],
        )

    if laser_evm.strategy.results:
        for addr in jump_addresses:
            output_list.append(laser_evm.strategy.results[addr])
    return output_list


def concolic_execution(
    concrete_data: ConcreteData, jump_addresses: List, solver_timeout=100000
) -> List[Dict[str, Dict[str, Any]]]:
    """
    Executes codes and prints input required to cover the branch flips
    :param input_file: Input file
    :param jump_addresses: Jump addresses to flip
    :param solver_timeout: Solver timeout

    """

    init_state, trace = concrete_execution(concrete_data)
    args.solver_timeout = solver_timeout
    output_list = flip_branches(
        init_state=init_state,
        concrete_data=concrete_data,
        jump_addresses=jump_addresses,
        trace=trace,
    )
    return output_list



"""This module contains detection code to find occurrences of calls whose
return value remains unchecked."""
from copy import copy
from typing import cast, List

from mythril.analysis import solver
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import UNCHECKED_RET_VAL
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.exceptions import UnsatError
from mythril.laser.smt import And
from mythril.laser.smt.bitvec import BitVec
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum.state.global_state import GlobalState

import logging
from typing_extensions import TypedDict

log = logging.getLogger(__name__)


class RetVal(TypedDict):
    address: int
    retval: BitVec


class UncheckedRetvalAnnotation(StateAnnotation):
    def __init__(self) -> None:
        self.retvals: List[RetVal] = []

    def __copy__(self):
        result = UncheckedRetvalAnnotation()
        result.retvals = copy(self.retvals)
        return result


class UncheckedRetval(DetectionModule):
    """A detection module to test whether CALL return value is checked."""

    name = "Return value of an external call is not checked"
    swc_id = UNCHECKED_RET_VAL
    description = (
        "Test whether CALL return value is checked. "
        "For direct calls, the Solidity compiler auto-generates this check. E.g.:\n"
        "    Alice c = Alice(address);\n"
        "    c.ping(42);\n"
        "Here the CALL will be followed by IZSERO(retval), if retval = ZERO then state is reverted. "
        "For low-level-calls this check is omitted. E.g.:\n"
        '    c.call.value(0)(bytes4(sha3("ping(uint256)")),1);'
    )
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["STOP", "RETURN"]
    post_hooks = ["CALL", "DELEGATECALL", "STATICCALL", "CALLCODE"]

    def _execute(self, state: GlobalState) -> List[Issue]:
        """

        :param state:
        :return:
        """
        return self._analyze_state(state)

    def _analyze_state(self, state: GlobalState) -> list:
        instruction = state.get_current_instruction()

        annotations = cast(
            List[UncheckedRetvalAnnotation],
            [a for a in state.get_annotations(UncheckedRetvalAnnotation)],
        )
        if len(annotations) == 0:
            state.annotate(UncheckedRetvalAnnotation())
            annotations = cast(
                List[UncheckedRetvalAnnotation],
                [a for a in state.get_annotations(UncheckedRetvalAnnotation)],
            )

        retvals = annotations[0].retvals

        if instruction["opcode"] in ("STOP", "RETURN"):
            issues = []
            for retval in retvals:
                try:
                    """
                    To check whether retval is unconstrained we are checking it against retval = 0 and retval = 1
                    """
                    solver.get_transaction_sequence(
                        state, state.world_state.constraints + [retval["retval"] == 1]
                    )
                    transaction_sequence = solver.get_transaction_sequence(
                        state, state.world_state.constraints + [retval["retval"] == 0]
                    )
                except UnsatError:
                    continue

                description_tail = (
                    "External calls return a boolean value. If the callee halts with an exception, 'false' is "
                    "returned and execution continues in the caller. "
                    "The caller should check whether an exception happened and react accordingly to avoid unexpected behavior. "
                    "For example it is often desirable to wrap external calls in require() so the transaction is reverted if the call fails."
                )

                issue = Issue(
                    contract=state.environment.active_account.contract_name,
                    function_name=state.environment.active_function_name,
                    address=retval["address"],
                    bytecode=state.environment.code.bytecode,
                    title="Unchecked return value from external call.",
                    swc_id=UNCHECKED_RET_VAL,
                    severity="Medium",
                    description_head="The return value of a message call is not checked.",
                    description_tail=description_tail,
                    gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
                    transaction_sequence=transaction_sequence,
                )
                conditions = [
                    And(*(state.world_state.constraints + [retval["retval"] == 1])),
                    And(*(state.world_state.constraints + [retval["retval"] == 0])),
                ]

                state.annotate(
                    IssueAnnotation(conditions=conditions, issue=issue, detector=self)
                )

                issues.append(issue)

            return issues
        else:
            log.debug("End of call, extracting retval")

            if state.environment.code.instruction_list[state.mstate.pc - 1][
                "opcode"
            ] not in ["CALL", "DELEGATECALL", "STATICCALL", "CALLCODE"]:
                # Return is pointless with OOG. The pc does not get updated in such cases
                return []

            return_value = state.mstate.stack[-1]
            retvals.append(
                {"address": state.instruction["address"] - 1, "retval": return_value}
            )

        return []


detector = UncheckedRetval()



import z3

from mythril.laser.smt.model import Model
from mythril.laser.smt.bool import Bool
from mythril.laser.smt.solver.solver_statistics import stat_smt_query

from typing import Set, Tuple, Dict, List, cast


def _get_expr_variables(expression: z3.ExprRef) -> List[z3.ExprRef]:
    """
    Gets the variables that make up the current expression
    :param expression:
    :return:
    """
    result = []
    if not expression.children() and not isinstance(expression, z3.BitVecNumRef):
        result.append(expression)
    for child in expression.children():
        c_children = _get_expr_variables(child)
        result.extend(c_children)
    return result


class DependenceBucket:
    """Bucket object to contain a set of conditions that are dependent on each other"""

    def __init__(self, variables=None, conditions=None):
        """
        Initializes a DependenceBucket object
        :param variables: Variables contained in the conditions
        :param conditions: The conditions that are dependent on each other
        """
        self.variables = variables or []  # type: List[z3.ExprRef]
        self.conditions = conditions or []  # type: List[z3.ExprRef]


class DependenceMap:
    """DependenceMap object that maintains a set of dependence buckets, used to separate independent smt queries"""

    def __init__(self):
        """Initializes a DependenceMap object"""
        self.buckets = []  # type: List[DependenceBucket]
        self.variable_map = {}  # type: Dict[str, DependenceBucket]

    def add_condition(self, condition: z3.BoolRef) -> None:
        """
        Add condition to the dependence map
        :param condition: The condition that is to be added to the dependence map
        """
        variables = set(_get_expr_variables(condition))
        relevant_buckets = set()
        for variable in variables:
            try:
                bucket = self.variable_map[str(variable)]
                relevant_buckets.add(bucket)
            except KeyError:
                continue

        new_bucket = DependenceBucket(variables, [condition])
        self.buckets.append(new_bucket)

        if relevant_buckets:
            # Merge buckets, and rewrite variable map accordingly
            relevant_buckets.add(new_bucket)
            new_bucket = self._merge_buckets(relevant_buckets)

        for variable in new_bucket.variables:
            self.variable_map[str(variable)] = new_bucket

    def _merge_buckets(self, bucket_list: Set[DependenceBucket]) -> DependenceBucket:
        """Merges the buckets in bucket list"""
        variables = []  # type: List[str]
        conditions = []  # type: List[z3.BoolRef]
        for bucket in bucket_list:
            self.buckets.remove(bucket)
            variables += bucket.variables
            conditions += bucket.conditions

        new_bucket = DependenceBucket(variables, conditions)
        self.buckets.append(new_bucket)

        return new_bucket


class IndependenceSolver:
    """An SMT solver object that uses independence optimization"""

    def __init__(self):
        """"""
        self.raw = z3.Solver()
        self.constraints = []
        self.models = []

    def set_timeout(self, timeout: int) -> None:
        """Sets the timeout that will be used by this solver, timeout is in
        milliseconds.

        :param timeout:
        """
        self.raw.set(timeout=timeout)

    def add(self, *constraints: Tuple[Bool]) -> None:
        """Adds the constraints to this solver.

        :param constraints: constraints to add
        """
        raw_constraints = [
            c.raw for c in cast(Tuple[Bool], constraints)
        ]  # type: List[z3.BoolRef]
        self.constraints.extend(raw_constraints)

    def append(self, *constraints: Tuple[Bool]) -> None:
        """Adds the constraints to this solver.

        :param constraints: constraints to add
        """
        raw_constraints = [
            c.raw for c in cast(Tuple[Bool], constraints)
        ]  # type: List[z3.BoolRef]
        self.constraints.extend(raw_constraints)

    @stat_smt_query
    def check(self) -> z3.CheckSatResult:
        """Returns z3 smt check result."""
        dependence_map = DependenceMap()
        for constraint in self.constraints:
            dependence_map.add_condition(constraint)

        self.models = []
        for bucket in dependence_map.buckets:
            self.raw.reset()
            self.raw.append(*bucket.conditions)
            check_result = self.raw.check()
            if check_result == z3.sat:
                self.models.append(self.raw.model())
            else:
                return check_result

        return z3.sat

    def model(self) -> Model:
        """Returns z3 model for a solution."""
        return Model(self.models)

    def reset(self) -> None:
        """Reset this solver."""
        self.constraints = []

    def pop(self, num) -> None:
        """Pop num constraints from this solver."""
        self.constraints.pop(num)



from mythril.support.support_utils import ModelCache
from mythril.support.support_args import args
from mythril.laser.smt import Optimize, simplify, And
from mythril.laser.ethereum.time_handler import time_handler
from mythril.exceptions import UnsatError, SolverTimeOutException

import logging
import os
import signal
import sys
import threading

from collections import OrderedDict
from copy import deepcopy
from functools import lru_cache
from multiprocessing.pool import ThreadPool
from multiprocessing import TimeoutError
from pathlib import Path
from time import time
from z3 import sat, unknown, is_true

log = logging.getLogger(__name__)


model_cache = ModelCache()


def solver_worker(
    constraints,
    minimize=(),
    maximize=(),
    solver_timeout=None,
):
    """
    Returns a model based on given constraints as a tuple
    :param constraints: Tuple of constraints
    :param minimize: Tuple of minimization conditions
    :param maximize: Tuple of maximization conditions
    :param solver_timeout: The timeout for solver
    :return:
    """
    s = Optimize()
    s.set_timeout(solver_timeout)

    for constraint in constraints:
        s.add(constraint)
    for e in minimize:
        s.minimize(e)
    for e in maximize:
        s.maximize(e)
    if args.solver_log:
        Path(args.solver_log).mkdir(parents=True, exist_ok=True)
        constraint_hash_input = tuple(
            list(constraints)
            + list(minimize)
            + list(maximize)
            + [len(constraints), len(minimize), len(maximize)]
        )
        with open(
            args.solver_log + f"/{abs(hash(constraint_hash_input))}.smt2", "w"
        ) as f:
            f.write(s.sexpr())

    result = s.check()
    return result, s


@lru_cache(maxsize=2**23)
def get_model(
    constraints,
    minimize=(),
    maximize=(),
    solver_timeout=None,
):
    """
    Returns a model based on given constraints as a tuple
    :param constraints: Tuple of constraints
    :param minimize: Tuple of minimization conditions
    :param maximize: Tuple of maximization conditions
    :param solver_timeout: The solver timeout
    :return:
    """

    solver_timeout = solver_timeout or args.solver_timeout
    solver_timeout = min(solver_timeout, time_handler.time_remaining())
    if solver_timeout <= 0:
        raise SolverTimeOutException
    for constraint in constraints:
        if type(constraint) == bool and not constraint:
            raise UnsatError

    if type(constraints) != tuple:
        constraints = constraints.get_all_constraints()
    constraints = [constraint for constraint in constraints if type(constraint) != bool]

    if len(maximize) + len(minimize) == 0:
        ret_model = model_cache.check_quick_sat(simplify(And(*constraints)).raw)
        if ret_model:
            return ret_model
    pool = ThreadPool(1)
    try:
        thread_result = pool.apply_async(
            solver_worker, args=(constraints, minimize, maximize, solver_timeout)
        )
        try:
            result, s = thread_result.get(solver_timeout)
        except TimeoutError:
            result = unknown
        except Exception:
            log.warning("Encountered an exception while solving expression using z3")
            result = unknown
    finally:
        # This is to prevent any segmentation faults from being displayed from z3
        sys.stdout = open(os.devnull, "w")
        sys.stderr = open(os.devnull, "w")
        pool.terminate()
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    if result == sat:
        model_cache.model_cache.put(s.model(), 1)
        return s.model()
    elif result == unknown:
        log.debug("Timeout/Error encountered while solving expression using z3")
        raise SolverTimeOutException
    raise UnsatError



"""This module contains the detection code for integer overflows and
underflows."""

from math import log2, ceil
from typing import cast, List, Set
from mythril.analysis import solver
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import INTEGER_OVERFLOW_AND_UNDERFLOW
from mythril.exceptions import UnsatError
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.analysis.module.base import DetectionModule, EntryPoint
from copy import copy

from mythril.laser.smt import (
    BVAddNoOverflow,
    BVSubNoUnderflow,
    BVMulNoOverflow,
    BitVec,
    If,
    symbol_factory,
    Not,
    Expression,
    Bool,
    And,
)

import logging

log = logging.getLogger(__name__)


class OverUnderflowAnnotation:
    """Symbol Annotation used if a BitVector can overflow"""

    def __init__(
        self, overflowing_state: GlobalState, operator: str, constraint: Bool
    ) -> None:
        self.overflowing_state = overflowing_state
        self.operator = operator
        self.constraint = constraint

    def __deepcopy__(self, memodict={}):
        new_annotation = copy(self)
        return new_annotation


class OverUnderflowStateAnnotation(StateAnnotation):
    """State Annotation used if an overflow is both possible and used in the annotated path"""

    def __init__(self) -> None:
        self.overflowing_state_annotations = set()  # type: Set[OverUnderflowAnnotation]

    def __copy__(self):
        new_annotation = OverUnderflowStateAnnotation()

        new_annotation.overflowing_state_annotations = copy(
            self.overflowing_state_annotations
        )

        return new_annotation


class IntegerArithmetics(DetectionModule):
    """This module searches for integer over- and underflows."""

    name = "Integer overflow or underflow"
    swc_id = INTEGER_OVERFLOW_AND_UNDERFLOW
    description = (
        "For every SUB instruction, check if there's a possible state "
        "where op1 > op0. For every ADD, MUL instruction, check if "
        "there's a possible state where op1 + op0 > 2^32 - 1"
    )
    entry_point = EntryPoint.CALLBACK
    pre_hooks = [
        "ADD",
        "MUL",
        "EXP",
        "SUB",
        "SSTORE",
        "JUMPI",
        "STOP",
        "RETURN",
        "CALL",
    ]

    def __init__(self) -> None:
        """
        Cache satisfiability of overflow constraints
        """
        super().__init__()

        self._ostates_satisfiable = set()  # type: Set[GlobalState]
        self._ostates_unsatisfiable = set()  # type: Set[GlobalState]

    def reset_module(self):
        """
        Resets the module
        :return:
        """
        super().reset_module()
        self._ostates_satisfiable = set()
        self._ostates_unsatisfiable = set()

    def _execute(self, state: GlobalState) -> List[Issue]:
        """Executes analysis module for integer underflow and integer overflow.

        :param state: Statespace to analyse
        :return: Found issues
        """

        opcode = state.get_current_instruction()["opcode"]

        funcs = {
            "ADD": [self._handle_add],
            "SUB": [self._handle_sub],
            "MUL": [self._handle_mul],
            "SSTORE": [self._handle_sstore],
            "JUMPI": [self._handle_jumpi],
            "CALL": [self._handle_call],
            "RETURN": [self._handle_return, self._handle_transaction_end],
            "STOP": [self._handle_transaction_end],
            "EXP": [self._handle_exp],
        }
        results = []
        for func in funcs[opcode]:
            result = func(state)
            if result and len(result) > 0:
                results += result
        return results

    def _get_args(self, state):
        stack = state.mstate.stack
        op0, op1 = (
            self._make_bitvec_if_not(stack, -1),
            self._make_bitvec_if_not(stack, -2),
        )
        return op0, op1

    def _handle_add(self, state):
        op0, op1 = self._get_args(state)
        c = Not(BVAddNoOverflow(op0, op1, False))

        annotation = OverUnderflowAnnotation(state, "addition", c)
        op0.annotate(annotation)

    def _handle_mul(self, state):
        op0, op1 = self._get_args(state)
        c = Not(BVMulNoOverflow(op0, op1, False))

        annotation = OverUnderflowAnnotation(state, "multiplication", c)
        op0.annotate(annotation)

    def _handle_sub(self, state):
        op0, op1 = self._get_args(state)
        c = Not(BVSubNoUnderflow(op0, op1, False))

        annotation = OverUnderflowAnnotation(state, "subtraction", c)
        op0.annotate(annotation)

    def _handle_exp(self, state):
        op0, op1 = self._get_args(state)

        if (op1.symbolic is False and op1.value == 0) or (
            op0.symbolic is False and op0.value < 2
        ):
            return
        if op0.symbolic and op1.symbolic:
            constraint = And(
                op1 > symbol_factory.BitVecVal(256, 256),
                op0 > symbol_factory.BitVecVal(1, 256),
            )
        elif op0.symbolic:
            constraint = op0 >= symbol_factory.BitVecVal(
                2 ** ceil(256 / op1.value), 256
            )
        else:
            constraint = op1 >= symbol_factory.BitVecVal(
                ceil(256 / log2(op0.value)), 256
            )

        annotation = OverUnderflowAnnotation(state, "exponentiation", constraint)
        op0.annotate(annotation)

    @staticmethod
    def _make_bitvec_if_not(stack, index):
        value = stack[index]
        if isinstance(value, BitVec):
            return value
        if isinstance(value, Bool):
            return If(value, 1, 0)
        stack[index] = symbol_factory.BitVecVal(value, 256)
        return stack[index]

    @staticmethod
    def _get_title(_type):
        return "Integer {}".format(_type)

    @staticmethod
    def _handle_sstore(state: GlobalState) -> None:

        stack = state.mstate.stack
        value = stack[-2]

        if not isinstance(value, Expression):
            return

        state_annotation = _get_overflowunderflow_state_annotation(state)
        for annotation in value.annotations:
            if isinstance(annotation, OverUnderflowAnnotation):
                state_annotation.overflowing_state_annotations.add(annotation)

    @staticmethod
    def _handle_jumpi(state):

        stack = state.mstate.stack
        value = stack[-2]

        state_annotation = _get_overflowunderflow_state_annotation(state)

        for annotation in value.annotations:
            if isinstance(annotation, OverUnderflowAnnotation):
                state_annotation.overflowing_state_annotations.add(annotation)

    @staticmethod
    def _handle_call(state):

        stack = state.mstate.stack
        value = stack[-3]

        state_annotation = _get_overflowunderflow_state_annotation(state)

        for annotation in value.annotations:
            if isinstance(annotation, OverUnderflowAnnotation):
                state_annotation.overflowing_state_annotations.add(annotation)

    @staticmethod
    def _handle_return(state: GlobalState) -> None:
        """
        Adds all the annotations into the state which correspond to the
        locations in the memory returned by RETURN opcode.
        :param state: The Global State
        """

        stack = state.mstate.stack
        offset, length = stack[-1], stack[-2]

        state_annotation = _get_overflowunderflow_state_annotation(state)

        for element in state.mstate.memory[offset : offset + length]:

            if not isinstance(element, Expression):
                continue

            for annotation in element.annotations:
                if isinstance(annotation, OverUnderflowAnnotation):
                    state_annotation.overflowing_state_annotations.add(annotation)

    def _handle_transaction_end(self, state: GlobalState) -> List[Issue]:

        state_annotation = _get_overflowunderflow_state_annotation(state)
        issues = []
        for annotation in state_annotation.overflowing_state_annotations:

            ostate = annotation.overflowing_state

            if ostate in self._ostates_unsatisfiable:
                continue

            if ostate not in self._ostates_satisfiable:
                try:
                    constraints = ostate.world_state.constraints + [
                        annotation.constraint
                    ]
                    solver.get_model(constraints)
                    self._ostates_satisfiable.add(ostate)
                except:
                    self._ostates_unsatisfiable.add(ostate)
                    continue

            log.debug(
                "Checking overflow in {} at transaction end address {}, ostate address {}".format(
                    state.get_current_instruction()["opcode"],
                    state.get_current_instruction()["address"],
                    ostate.get_current_instruction()["address"],
                )
            )

            try:

                constraints = state.world_state.constraints + [annotation.constraint]
                transaction_sequence = solver.get_transaction_sequence(
                    state, constraints
                )
            except UnsatError:
                continue

            description_head = "The arithmetic operator can {}.".format(
                "underflow" if annotation.operator == "subtraction" else "overflow"
            )
            description_tail = "It is possible to cause an integer overflow or underflow in the arithmetic operation. "
            "Prevent this by constraining inputs using the require() statement or use the OpenZeppelin "
            "SafeMath library for integer arithmetic operations. "
            "Refer to the transaction trace generated for this issue to reproduce the issue."

            issue = Issue(
                contract=ostate.environment.active_account.contract_name,
                function_name=ostate.environment.active_function_name,
                address=ostate.get_current_instruction()["address"],
                swc_id=INTEGER_OVERFLOW_AND_UNDERFLOW,
                bytecode=ostate.environment.code.bytecode,
                title="Integer Arithmetic Bugs",
                severity="High",
                description_head=description_head,
                description_tail=description_tail,
                gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
                transaction_sequence=transaction_sequence,
            )
            state.annotate(
                IssueAnnotation(
                    issue=issue, detector=self, conditions=[And(*constraints)]
                )
            )
            issues.append(issue)
        return issues


detector = IntegerArithmetics()


def _get_address_from_state(state):
    return state.get_current_instruction()["address"]


def _get_overflowunderflow_state_annotation(
    state: GlobalState,
) -> OverUnderflowStateAnnotation:
    state_annotations = cast(
        List[OverUnderflowStateAnnotation],
        list(state.get_annotations(OverUnderflowStateAnnotation)),
    )

    if len(state_annotations) == 0:
        state_annotation = OverUnderflowStateAnnotation()
        state.annotate(state_annotation)
        return state_annotation
    else:
        return state_annotations[0]



"""This module contains the detection code for Arbitrary jumps."""
import logging

from mythril.exceptions import UnsatError

from mythril.analysis.solver import get_transaction_sequence, UnsatError
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.module.base import DetectionModule, Issue, EntryPoint
from mythril.analysis.swc_data import ARBITRARY_JUMP

from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import And, BitVec, symbol_factory
from mythril.support.model import get_model

log = logging.getLogger(__name__)

DESCRIPTION = """

Search for jumps to arbitrary locations in the bytecode
"""


def is_unique_jumpdest(jump_dest: BitVec, state: GlobalState) -> bool:
    """
    Handles cases where jump_dest evaluates to a single concrete value
    """

    try:
        model = get_model(state.world_state.constraints)
    except UnsatError:
        return True
    concrete_jump_dest = model.eval(jump_dest.raw, model_completion=True)
    try:
        model = get_model(
            state.world_state.constraints
            + [symbol_factory.BitVecVal(concrete_jump_dest.as_long(), 256) != jump_dest]
        )
    except UnsatError:
        return True
    return False


class ArbitraryJump(DetectionModule):
    """This module searches for JUMPs to a user-specified location."""

    name = "Caller can redirect execution to arbitrary bytecode locations"
    swc_id = ARBITRARY_JUMP
    description = DESCRIPTION
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["JUMP", "JUMPI"]

    def reset_module(self):
        """
        Resets the module by clearing everything
        :return:
        """
        super().reset_module()

    def _execute(self, state: GlobalState) -> None:
        """

        :param state:
        :return:
        """
        return self._analyze_state(state)

    def _analyze_state(self, state):
        """

        :param state:
        :return:
        """

        jump_dest = state.mstate.stack[-1]

        if jump_dest.symbolic is False:
            return []

        if is_unique_jumpdest(jump_dest, state) is True:
            return []

        try:
            transaction_sequence = get_transaction_sequence(
                state, state.world_state.constraints
            )
        except UnsatError:
            return []
        log.info("Detected arbitrary jump dest")
        issue = Issue(
            contract=state.environment.active_account.contract_name,
            function_name=state.environment.active_function_name,
            address=state.get_current_instruction()["address"],
            swc_id=ARBITRARY_JUMP,
            title="Jump to an arbitrary instruction",
            severity="High",
            bytecode=state.environment.code.bytecode,
            description_head="The caller can redirect execution to arbitrary bytecode locations.",
            description_tail="It is possible to redirect the control flow to arbitrary locations in the code. "
            "This may allow an attacker to bypass security controls or manipulate the business logic of the "
            "smart contract. Avoid using low-level-operations and assembly to prevent this issue.",
            gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
            transaction_sequence=transaction_sequence,
        )
        state.annotate(
            IssueAnnotation(
                conditions=[And(*state.world_state.constraints)],
                issue=issue,
                detector=self,
            )
        )

        return [issue]


detector = ArbitraryJump()



"""This module provides helper functions for the analysis modules to deal with
call functionality."""
from typing import Union

from mythril.analysis.ops import VarType, Call, get_variable
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.natives import PRECOMPILE_COUNT


def get_call_from_state(state: GlobalState) -> Union[Call, None]:
    """

    :param state:
    :return:
    """
    instruction = state.get_current_instruction()

    op = instruction["opcode"]
    stack = state.mstate.stack

    if op in ("CALL", "CALLCODE"):
        gas, to, value, meminstart, meminsz, _, _ = (
            get_variable(stack[-1]),
            get_variable(stack[-2]),
            get_variable(stack[-3]),
            get_variable(stack[-4]),
            get_variable(stack[-5]),
            get_variable(stack[-6]),
            get_variable(stack[-7]),
        )

        if to.type == VarType.CONCRETE and 0 < to.val <= PRECOMPILE_COUNT:
            return None

        if meminstart.type == VarType.CONCRETE and meminsz.type == VarType.CONCRETE:
            return Call(
                state.node,
                state,
                None,
                op,
                to,
                gas,
                value,
                state.mstate.memory[meminstart.val : meminsz.val * 4],
            )
        else:
            return Call(state.node, state, None, op, to, gas, value)

    else:
        gas, to, meminstart, meminsz, _, _ = (
            get_variable(stack[-1]),
            get_variable(stack[-2]),
            get_variable(stack[-3]),
            get_variable(stack[-4]),
            get_variable(stack[-5]),
            get_variable(stack[-6]),
        )

        return Call(state.node, state, None, op, to, gas)



"""This module contains functions setting up and executing transactions with
symbolic values."""
import logging
from typing import Optional, List, Union
from copy import deepcopy


from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.cfg import Node, Edge, JumpType
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.calldata import SymbolicCalldata
from mythril.laser.ethereum.state.constraints import Constraints
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.transaction.transaction_models import (
    MessageCallTransaction,
    ContractCreationTransaction,
    tx_id_manager,
    BaseTransaction,
)
from mythril.laser.smt import symbol_factory, Or, Bool, BitVec
from mythril.support.support_args import args as cmd_args


FUNCTION_HASH_BYTE_LENGTH = 4

log = logging.getLogger(__name__)


class Actors:
    def __init__(
        self,
        creator=0xAFFEAFFEAFFEAFFEAFFEAFFEAFFEAFFEAFFEAFFE,
        attacker=0xDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF,
        someguy=0xAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,
    ):
        self.addresses = {
            "CREATOR": symbol_factory.BitVecVal(creator, 256),
            "ATTACKER": symbol_factory.BitVecVal(attacker, 256),
            "SOMEGUY": symbol_factory.BitVecVal(someguy, 256),
        }

    def __setitem__(self, actor: str, address: Optional[str]):
        """
        Sets an actor to a desired address

        :param actor: Name of the actor to set
        :param address: Address to set the actor to. None to delete the actor
        """
        if address is None:
            if actor in ("CREATOR", "ATTACKER"):
                raise ValueError("Can't delete creator or attacker address")
            del self.addresses[actor]
        else:
            if address[0:2] != "0x":
                raise ValueError("Actor address not in valid format")

            self.addresses[actor] = symbol_factory.BitVecVal(int(address[2:], 16), 256)

    def __getitem__(self, actor: str):
        return self.addresses[actor]

    @property
    def creator(self):
        return self.addresses["CREATOR"]

    @property
    def attacker(self):
        return self.addresses["ATTACKER"]

    def __len__(self):
        return len(self.addresses)


ACTORS = Actors()


def generate_function_constraints(
    calldata: SymbolicCalldata, func_hashes: List[List[int]]
) -> List[Bool]:
    """
    This will generate constraints for fixing the function call part of calldata
    :param calldata: Calldata
    :param func_hashes: The list of function hashes allowed for this transaction
    :return: Constraints List
    """
    if len(func_hashes) == 0:
        return []
    constraints = []
    for i in range(FUNCTION_HASH_BYTE_LENGTH):
        constraint = Bool(False)
        for func_hash in func_hashes:
            if func_hash == -1:
                # Fallback function without calldata
                constraint = Or(constraint, calldata.size < 4)
            elif func_hash == -2:
                # Receive function
                constraint = Or(constraint, calldata.size == 0)
            else:
                constraint = Or(
                    constraint, calldata[i] == symbol_factory.BitVecVal(func_hash[i], 8)
                )
        constraints.append(constraint)
    return constraints


def execute_message_call(
    laser_evm, callee_address: BitVec, func_hashes: List[List[int]] = None
) -> None:
    """Executes a message call transaction from all open states.

    :param laser_evm:
    :param callee_address:
    """
    # TODO: Resolve circular import between .transaction and ..svm to import LaserEVM here
    open_states = laser_evm.open_states[:]
    del laser_evm.open_states[:]

    for open_world_state in open_states:
        if open_world_state[callee_address].deleted:
            log.debug("Can not execute dead contract, skipping.")
            continue

        next_transaction_id = tx_id_manager.get_next_tx_id()

        external_sender = symbol_factory.BitVecSym(
            "sender_{}".format(next_transaction_id), 256
        )
        calldata = SymbolicCalldata(next_transaction_id)
        transaction = MessageCallTransaction(
            world_state=open_world_state,
            identifier=next_transaction_id,
            gas_price=symbol_factory.BitVecSym(
                "gas_price{}".format(next_transaction_id), 256
            ),
            gas_limit=8000000,  # block gas limit
            origin=external_sender,
            caller=external_sender,
            callee_account=open_world_state[callee_address],
            call_data=calldata,
            call_value=symbol_factory.BitVecSym(
                "call_value{}".format(next_transaction_id), 256
            ),
        )
        constraints = (
            generate_function_constraints(calldata, func_hashes)
            if func_hashes
            else None
        )
        _setup_global_state_for_execution(laser_evm, transaction, constraints)

    laser_evm.exec()


def execute_contract_creation(
    laser_evm,
    contract_initialization_code,
    contract_name=None,
    world_state=None,
    origin=ACTORS["CREATOR"],
    caller=ACTORS["CREATOR"],
) -> Account:
    """Executes a contract creation transaction from all open states.

    :param laser_evm:
    :param contract_initialization_code:
    :param contract_name:
    :return:
    """

    world_state = world_state or WorldState()
    open_states = [world_state]
    del laser_evm.open_states[:]
    new_account = None
    for open_world_state in open_states:
        next_transaction_id = tx_id_manager.get_next_tx_id()
        # call_data "should" be '[]', but it is easier to model the calldata symbolically
        # and add logic in codecopy/codesize/calldatacopy/calldatasize than to model code "correctly"
        transaction = ContractCreationTransaction(
            world_state=open_world_state,
            identifier=next_transaction_id,
            gas_price=symbol_factory.BitVecSym(
                "gas_price{}".format(next_transaction_id), 256
            ),
            gas_limit=8000000,  # block gas limit
            origin=origin,
            code=Disassembly(contract_initialization_code),
            caller=caller,
            contract_name=contract_name,
            call_data=None,
            call_value=symbol_factory.BitVecSym(
                "call_value{}".format(next_transaction_id), 256
            ),
        )
        _setup_global_state_for_execution(laser_evm, transaction)
        new_account = new_account or transaction.callee_account

    laser_evm.exec(True)

    return new_account


def _setup_global_state_for_execution(
    laser_evm,
    transaction: BaseTransaction,
    initial_constraints: Optional[List[Bool]] = None,
) -> None:
    """Sets up global state and cfg for a transactions execution.

    :param laser_evm:
    :param transaction:
    """
    # TODO: Resolve circular import between .transaction and ..svm to import LaserEVM here
    global_state = transaction.initial_global_state()
    global_state.transaction_stack.append((transaction, None))
    global_state.world_state.constraints += initial_constraints or []

    global_state.world_state.constraints.append(
        Or(*[transaction.caller == actor for actor in ACTORS.addresses.values()])
    )

    new_node = Node(
        global_state.environment.active_account.contract_name,
        function_name=global_state.environment.active_function_name,
    )
    if laser_evm.requires_statespace:
        laser_evm.nodes[new_node.uid] = new_node

    if transaction.world_state.node:
        if laser_evm.requires_statespace:
            laser_evm.edges.append(
                Edge(
                    transaction.world_state.node.uid,
                    new_node.uid,
                    edge_type=JumpType.Transaction,
                    condition=None,
                )
            )
        new_node.constraints = global_state.world_state.constraints

    global_state.world_state.transaction_sequence.append(transaction)
    global_state.node = new_node
    new_node.states.append(global_state)
    laser_evm.work_list.append(global_state)


def execute_transaction(*args, **kwargs):
    """
    Chooses the transaction type based on callee address and
    executes the transaction
    """
    laser_evm = args[0]
    if kwargs["callee_address"] == "":
        for ws in laser_evm.open_states[:]:
            execute_contract_creation(
                laser_evm=laser_evm,
                contract_initialization_code=kwargs["data"],
                world_state=ws,
            )
        return

    execute_message_call(
        laser_evm=laser_evm,
        callee_address=symbol_factory.BitVecVal(int(kwargs["callee_address"], 16), 256),
    )



""" Laser plugins

This module contains everything to do with laser plugins

Laser plugins are a way of extending laser's functionality without complicating the core business logic.
Different features that have been implemented in the form of plugins are:
- benchmarking
- path pruning

Plugins also provide a way to implement optimisations outside of the mythril code base and to inject them.
The api that laser currently provides is still unstable and will probably change to suit our needs
as more plugins get developed.

For the implementation of plugins the following modules are of interest:
- laser.plugins.plugin
- laser.plugins.signals
- laser.svm

Which show the basic interfaces with which plugins are able to interact
"""



"""This module implements basic symbolic execution search strategies."""
from random import randrange
from typing import List

from mythril.laser.ethereum.state.global_state import GlobalState
from . import BasicSearchStrategy
from random import choices


class DepthFirstSearchStrategy(BasicSearchStrategy):
    """Implements a depth first search strategy I.E.

    Follow one path to a leaf, and then continue to the next one
    """

    def get_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        return self.work_list.pop()

    def view_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        return self.work_list[-1]


class BreadthFirstSearchStrategy(BasicSearchStrategy):
    """Implements a breadth first search strategy I.E.

    Execute all states of a "level" before continuing
    """

    def get_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        return self.work_list.pop(0)

    def view_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        return self.work_list[0]


class ReturnRandomNaivelyStrategy(BasicSearchStrategy):
    """chooses a random state from the worklist with equal likelihood."""

    def __init__(self, work_list, max_depth, **kwargs):
        super().__init__(work_list, max_depth, **kwargs)
        self.previous_random_value = -1

    def get_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        if len(self.work_list) > 0:
            if self.previous_random_value == -1:
                return self.work_list.pop(randrange(len(self.work_list)))
            else:
                new_state = self.work_list.pop(self.previous_random_value)
                self.previous_random_value = -1
                return new_state
        else:
            raise IndexError

    def view_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        if len(self.work_list) > 0:
            self.previous_random_value = randrange(len(self.work_list))
            return self.work_list[self.previous_random_value]
        else:
            raise IndexError


class ReturnWeightedRandomStrategy(BasicSearchStrategy):
    """chooses a random state from the worklist with likelihood based on
    inverse proportion to depth."""

    def __init__(self, work_list, max_depth, **kwargs):
        super().__init__(work_list, max_depth, **kwargs)
        self.previous_random_value = -1

    def get_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        probability_distribution = [
            1 / (global_state.mstate.depth + 1) for global_state in self.work_list
        ]
        if self.previous_random_value != -1:
            ns = self.work_list.pop(self.previous_random_value)
            self.previous_random_value = -1
            return ns
        else:
            return self.work_list.pop(
                choices(range(len(self.work_list)), probability_distribution)[0]
            )

    def view_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        probability_distribution = [
            1 / (global_state.mstate.depth + 1) for global_state in self.work_list
        ]
        self.previous_random_value = choices(
            range(len(self.work_list)), probability_distribution
        )[0]
        return self.work_list[self.previous_random_value]



import z3

from typing import Union, List


class Model:
    """The model class wraps a z3 model

    This implementation allows for multiple internal models, this is required for the use of an independence solver
    which has models for multiple queries which need an uniform output.
    """

    def __init__(self, models: List[z3.ModelRef] = None):
        """
        Initialize a model object
        :param models: the internal z3 models that this model should use
        """
        self.raw = models or []

    def decls(self) -> List[z3.ExprRef]:
        """Get the declarations for this model"""
        result = []  # type: List[z3.ExprRef]
        for internal_model in self.raw:
            result.extend(internal_model.decls())
        return result

    def __getitem__(self, item) -> Union[None, z3.ExprRef]:
        """Get declaration from model
        If item is an int, then the declaration at offset item is returned
        If item is a declaration, then the interpretation is returned
        """
        for internal_model in self.raw:
            is_last_model = self.raw.index(internal_model) == len(self.raw) - 1

            try:
                result = internal_model[item]
                if result is not None:
                    return result
            except IndexError:
                if is_last_model:
                    raise
                continue
        return None

    def eval(
        self, expression: z3.ExprRef, model_completion: bool = False
    ) -> Union[None, z3.ExprRef]:
        """Evaluate the expression using this model

        :param expression: The expression to evaluate
        :param model_completion: Use the default value if the model has no interpretation of the given expression
        :return: The evaluated expression
        """
        for internal_model in self.raw:
            is_last_model = self.raw.index(internal_model) == len(self.raw) - 1
            is_relevant_model = expression.decl() in list(internal_model.decls())
            if is_relevant_model or is_last_model:
                return internal_model.eval(expression, model_completion)
        return None



"""This module contains the SMT abstraction for a basic symbol expression."""
from typing import Optional, Set, Any, TypeVar, Generic, cast
import z3


Annotations = Set[Any]
T = TypeVar("T", bound=z3.ExprRef)


class Expression(Generic[T]):
    """This is the base symbol class and maintains functionality for
    simplification and annotations."""

    def __init__(self, raw: T, annotations: Optional[Annotations] = None):
        """

        :param raw:
        :param annotations:
        """
        self.raw = raw

        if annotations:
            assert isinstance(annotations, set)

        self._annotations = annotations or set()

    @property
    def annotations(self) -> Annotations:
        """Gets the annotations for this expression.

        :return:
        """

        return self._annotations

    def annotate(self, annotation: Any) -> None:
        """Annotates this expression with the given annotation.

        :param annotation:
        """

        self._annotations.add(annotation)

    def simplify(self) -> None:
        """Simplify this expression."""
        self.raw = cast(T, z3.simplify(self.raw))

    def __repr__(self) -> str:
        return repr(self.raw)

    def size(self):
        return self.raw.size()

    def __hash__(self) -> int:
        return self.raw.__hash__()

    def get_annotations(self, annotation: Any):
        return list(filter(lambda x: isinstance(x, annotation), self.annotations))


G = TypeVar("G", bound=Expression)


def simplify(expression: G) -> G:
    """Simplify the expression .

    :param expression:
    :return:
    """
    expression.simplify()
    return expression



from mythril.laser.smt.bitvec import BitVec

from mythril.laser.smt.bitvec_helper import (
    If,
    UGT,
    ULT,
    ULE,
    Concat,
    Extract,
    URem,
    SRem,
    UDiv,
    UGE,
    Sum,
    BVAddNoOverflow,
    BVMulNoOverflow,
    BVSubNoUnderflow,
    LShR,
)

from mythril.laser.smt.expression import Expression, simplify
from mythril.laser.smt.bool import Bool, is_true, is_false, Or, Not, And
from mythril.laser.smt.array import K, Array, BaseArray
from mythril.laser.smt.function import Function
from mythril.laser.smt.solver import Solver, Optimize, SolverStatistics
from mythril.laser.smt.model import Model
from mythril.laser.smt.bool import Bool as SMTBool
from typing import Union, Any, Optional, Set, TypeVar, Generic
import z3


Annotations = Optional[Set[Any]]
T = TypeVar("T", bound=Union[SMTBool, z3.BoolRef])
U = TypeVar("U", bound=Union[BitVec, z3.BitVecRef])


class SymbolFactory(Generic[T, U]):
    """A symbol factory provides a default interface for all the components of mythril to create symbols"""

    @staticmethod
    def Bool(value: "__builtins__.bool", annotations: Annotations = None) -> T:
        """
        Creates a Bool with concrete value
        :param value: The boolean value
        :param annotations: The annotations to initialize the bool with
        :return: The freshly created Bool()
        """
        raise NotImplementedError

    @staticmethod
    def BoolSym(name: str, annotations: Annotations = None) -> T:
        """
        Creates a boolean symbol
        :param name: The name of the Bool variable
        :param annotations: The annotations to initialize the bool with
        :return: The freshly created Bool()
        """
        raise NotImplementedError

    @staticmethod
    def BitVecVal(value: int, size: int, annotations: Annotations = None) -> U:
        """Creates a new bit vector with a concrete value.

        :param value: The concrete value to set the bit vector to
        :param size: The size of the bit vector
        :param annotations: The annotations to initialize the bit vector with
        :return: The freshly created bit vector
        """
        raise NotImplementedError()

    @staticmethod
    def BitVecSym(name: str, size: int, annotations: Annotations = None) -> U:
        """Creates a new bit vector with a symbolic value.

        :param name: The name of the symbolic bit vector
        :param size: The size of the bit vector
        :param annotations: The annotations to initialize the bit vector with
        :return: The freshly created bit vector
        """
        raise NotImplementedError()


class _SmtSymbolFactory(SymbolFactory[SMTBool, BitVec]):
    """
    An implementation of a SymbolFactory that creates symbols using
    the classes in: mythril.laser.smt
    """

    @staticmethod
    def Bool(value: "__builtins__.bool", annotations: Annotations = None) -> SMTBool:
        """
        Creates a Bool with concrete value
        :param value: The boolean value
        :param annotations: The annotations to initialize the bool with
        :return: The freshly created Bool()
        """
        raw = z3.BoolVal(value)
        return SMTBool(raw, annotations)

    @staticmethod
    def BoolSym(name: str, annotations: Annotations = None) -> SMTBool:
        """
        Creates a boolean symbol
        :param name: The name of the Bool variable
        :param annotations: The annotations to initialize the bool with
        :return: The freshly created Bool()
        """
        raw = z3.Bool(name)
        return SMTBool(raw, annotations)

    @staticmethod
    def BitVecVal(value: int, size: int, annotations: Annotations = None) -> BitVec:
        """Creates a new bit vector with a concrete value."""
        raw = z3.BitVecVal(value, size)
        return BitVec(raw, annotations)

    @staticmethod
    def BitVecSym(name: str, size: int, annotations: Annotations = None) -> BitVec:
        """Creates a new bit vector with a symbolic value."""
        raw = z3.BitVec(name, size)
        return BitVec(raw, annotations)


class _Z3SymbolFactory(SymbolFactory[z3.BoolRef, z3.BitVecRef]):
    """
    An implementation of a SymbolFactory that directly returns
    z3 symbols
    """

    @staticmethod
    def Bool(value: "__builtins__.bool", annotations: Annotations = None) -> z3.BoolRef:
        """Creates a new bit vector with a concrete value"""
        return z3.BoolVal(value)

    @staticmethod
    def BitVecVal(
        value: int, size: int, annotations: Annotations = None
    ) -> z3.BitVecRef:
        """Creates a new bit vector with a concrete value."""
        return z3.BitVecVal(value, size)

    @staticmethod
    def BitVecSym(
        name: str, size: int, annotations: Annotations = None
    ) -> z3.BitVecRef:
        """Creates a new bit vector with a symbolic value."""
        return z3.BitVec(name, size)


# This is the instance that other parts of mythril should use

# Type hints are not allowed here in 3.5
# symbol_factory: SymbolFactory = _SmtSymbolFactory()
symbol_factory = _SmtSymbolFactory()



"""This module contains the business logic used by Instruction in
instructions.py to get the necessary elements from the stack and determine the
parameters for the new global state."""

import logging
import re
from typing import Union, List, cast, Optional
from eth.constants import GAS_CALLSTIPEND

import mythril.laser.ethereum.util as util
from mythril.laser.ethereum.util import insert_ret_val
from mythril.laser.ethereum import natives
from mythril.laser.ethereum.cheat_code import handle_cheat_codes, hevm_cheat_code
from mythril.laser.ethereum.instruction_data import calculate_native_gas
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.natives import PRECOMPILE_COUNT, PRECOMPILE_FUNCTIONS
from mythril.laser.ethereum.state.calldata import (
    BaseCalldata,
    SymbolicCalldata,
    ConcreteCalldata,
)
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import BitVec, If
from mythril.laser.smt import simplify, Expression, symbol_factory
from mythril.support.loader import DynLoader

"""
This module contains the business logic used by Instruction in instructions.py
to get the necessary elements from the stack and determine the parameters for the new global state.
"""

log = logging.getLogger(__name__)
SYMBOLIC_CALLDATA_SIZE = 320  # Used when copying symbolic calldata


def get_call_parameters(
    global_state: GlobalState, dynamic_loader: DynLoader, with_value=False
):
    """Gets call parameters from global state Pops the values from the stack
    and determines output parameters.


    :param global_state: state to look in
    :param dynamic_loader: dynamic loader to use
    :param with_value: whether to pop the value argument from the stack
    :return: callee_account, call_data, value, call_data_type, gas
    """
    gas, to = global_state.mstate.pop(2)
    value = global_state.mstate.pop() if with_value else 0
    (
        memory_input_offset,
        memory_input_size,
        memory_out_offset,
        memory_out_size,
    ) = global_state.mstate.pop(4)

    callee_address = get_callee_address(global_state, dynamic_loader, to)

    callee_account = None
    call_data = get_call_data(global_state, memory_input_offset, memory_input_size)
    if isinstance(callee_address, BitVec) or (
        isinstance(callee_address, str)
        and (int(callee_address, 16) > PRECOMPILE_COUNT or int(callee_address, 16) == 0)
    ):
        callee_account = get_callee_account(
            global_state, callee_address, dynamic_loader
        )

    gas = gas + If(value > 0, symbol_factory.BitVecVal(GAS_CALLSTIPEND, gas.size()), 0)
    return (
        callee_address,
        callee_account,
        call_data,
        value,
        gas,
        memory_out_offset,
        memory_out_size,
    )


def _get_padded_hex_address(address: int) -> str:
    hex_address = hex(address)[2:]
    return "0x{}{}".format("0" * (40 - len(hex_address)), hex_address)


def get_callee_address(
    global_state: GlobalState,
    dynamic_loader: DynLoader,
    symbolic_to_address: Expression,
):
    """Gets the address of the callee.

    :param global_state: state to look in
    :param dynamic_loader:  dynamic loader to use
    :param symbolic_to_address: The (symbolic) callee address
    :return: Address of the callee
    """
    environment = global_state.environment
    try:
        callee_address = _get_padded_hex_address(
            util.get_concrete_int(symbolic_to_address)
        )
    except TypeError:
        log.debug("Symbolic call encountered")

        match = re.search(r"Storage\[(\d+)\]", str(simplify(symbolic_to_address)))

        if match is None or dynamic_loader is None:
            return symbolic_to_address

        index = int(match.group(1))
        log.debug("Dynamic contract address at storage index {}".format(index))

        # attempt to read the contract address from instance storage
        try:
            callee_address = dynamic_loader.read_storage(
                "0x{:040X}".format(environment.active_account.address.value), index
            )
        # TODO: verify whether this happens or not
        except:
            return symbolic_to_address

        # testrpc simply returns the address, geth response is more elaborate.
        if not re.match(r"^0x[0-9a-f]{40}$", callee_address):
            callee_address = "0x" + callee_address[26:]

    return callee_address


def get_callee_account(
    global_state: GlobalState,
    callee_address: Union[str, BitVec],
    dynamic_loader: DynLoader,
):
    """Gets the callees account from the global_state.

    :param global_state: state to look in
    :param callee_address: address of the callee
    :param dynamic_loader: dynamic loader to use
    :return: Account belonging to callee
    """
    if isinstance(callee_address, BitVec):
        if callee_address.symbolic:
            return Account(callee_address, balances=global_state.world_state.balances)
        else:
            callee_address = hex(callee_address.value)[2:]

    return global_state.world_state.accounts_exist_or_load(
        callee_address, dynamic_loader
    )


def get_call_data(
    global_state: GlobalState,
    memory_start: Union[int, BitVec],
    memory_size: Union[int, BitVec],
):
    """Gets call_data from the global_state.

    :param global_state: state to look in
    :param memory_start: Start index
    :param memory_size: Size
    :return: Tuple containing: call_data array from memory or empty array if symbolic, type found
    """
    state = global_state.mstate
    transaction_id = "{}_internalcall".format(global_state.current_transaction.id)

    memory_start = cast(
        BitVec,
        (
            symbol_factory.BitVecVal(memory_start, 256)
            if isinstance(memory_start, int)
            else memory_start
        ),
    )
    memory_size = cast(
        BitVec,
        (
            symbol_factory.BitVecVal(memory_size, 256)
            if isinstance(memory_size, int)
            else memory_size
        ),
    )

    if memory_size.symbolic:
        memory_size = SYMBOLIC_CALLDATA_SIZE
    try:
        calldata_from_mem = state.memory[
            util.get_concrete_int(memory_start) : util.get_concrete_int(
                memory_start + memory_size
            )
        ]
        return ConcreteCalldata(transaction_id, calldata_from_mem)
    except TypeError:
        log.debug("Unsupported symbolic memory offset and size")
        return SymbolicCalldata(transaction_id)


def native_call(
    global_state: GlobalState,
    callee_address: Union[str, BitVec],
    call_data: BaseCalldata,
    memory_out_offset: Union[int, Expression],
    memory_out_size: Union[int, Expression],
) -> Optional[List[GlobalState]]:

    if isinstance(callee_address, BitVec) or not (
        0 < int(callee_address, 16) <= PRECOMPILE_COUNT
        or hevm_cheat_code.is_cheat_address(callee_address)
    ):
        return None
    if hevm_cheat_code.is_cheat_address(callee_address):
        log.info("HEVM cheat code address triggered")
        handle_cheat_codes(
            global_state, callee_address, call_data, memory_out_offset, memory_out_size
        )
        return [global_state]

    log.debug("Native contract called: " + callee_address)
    try:
        mem_out_start = util.get_concrete_int(memory_out_offset)
        mem_out_sz = util.get_concrete_int(memory_out_size)
    except TypeError:
        insert_ret_val(global_state)
        log.debug("CALL with symbolic start or offset not supported")
        return [global_state]

    call_address_int = int(callee_address, 16)
    native_gas_min, native_gas_max = calculate_native_gas(
        global_state.mstate.calculate_extension_size(mem_out_start, mem_out_sz),
        PRECOMPILE_FUNCTIONS[call_address_int - 1].__name__,
    )
    global_state.mstate.min_gas_used += native_gas_min
    global_state.mstate.max_gas_used += native_gas_max
    global_state.mstate.mem_extend(mem_out_start, mem_out_sz)

    try:
        data = natives.native_contracts(call_address_int, call_data)
    except natives.NativeContractException:
        for i in range(mem_out_sz):
            global_state.mstate.memory[mem_out_start + i] = global_state.new_bitvec(
                PRECOMPILE_FUNCTIONS[call_address_int - 1].__name__
                + "("
                + str(call_data)
                + ")",
                8,
            )
        insert_ret_val(global_state)
        return [global_state]

    for i in range(
        min(len(data), mem_out_sz)
    ):  # If more data is used then it's chopped off
        global_state.mstate.memory[mem_out_start + i] = data[i]

    insert_ret_val(global_state)
    return [global_state]



import traceback


def is_prehook() -> bool:
    """Check if we are in prehook.  One of Bernhard's trademark hacks!
    Let's leave it to this for now, unless we need to check prehook for
    a lot more modules.
    """

    assert ("pre_hook" in traceback.format_stack()[-5]) or (
        "post_hook" in traceback.format_stack()[-5]
    )
    return "pre_hook" in traceback.format_stack()[-5]



import pytest

from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.state.environment import Environment
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.machine_state import MachineState
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.instructions import Instruction
from mythril.laser.ethereum.transaction.transaction_models import MessageCallTransaction
from mythril.laser.smt import symbol_factory, simplify


def get_state(input):
    world_state = WorldState()
    account = world_state.create_account(balance=10, address=101)
    account.code = Disassembly(input)
    environment = Environment(account, None, None, None, None, None, None)
    state = GlobalState(world_state, environment, None, MachineState(gas_limit=8000000))
    state.transaction_stack.append(
        (MessageCallTransaction(world_state=WorldState(), gas_limit=8000000), None)
    )
    return state


BVV = symbol_factory.BitVecVal
BV = symbol_factory.BitVecSym

test_data = [
    ("5F", 0),  # Successful execution, stack consists of a single item, set to zero
    (
        "5F" * 1024,
        0,
    ),  # Successful execution, stack consists of 1024 items, all set to zero
]


@pytest.mark.parametrize("inputs,output", test_data)
def test_push0(inputs, output):
    # Arrange
    state = get_state(inputs)

    instruction = Instruction("push", dynamic_loader=None)

    # Act
    new_state = instruction.evaluate(state)[0]

    # Assert
    assert simplify(new_state.mstate.stack[-1]) == output



"""This module contains the detection code for arbitrary storage write."""
import logging
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.analysis.potential_issues import (
    get_potential_issues_annotation,
    PotentialIssue,
)

from mythril.analysis.swc_data import WRITE_TO_ARBITRARY_STORAGE
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import symbol_factory

log = logging.getLogger(__name__)

DESCRIPTION = """

Search for any writes to an arbitrary storage slot
"""


class ArbitraryStorage(DetectionModule):
    """This module searches for a feasible write to an arbitrary storage slot."""

    name = "Caller can write to arbitrary storage locations"
    swc_id = WRITE_TO_ARBITRARY_STORAGE
    description = DESCRIPTION
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["SSTORE"]

    def reset_module(self):
        """
        Resets the module by clearing everything
        :return:
        """
        super().reset_module()

    def _execute(self, state: GlobalState) -> None:
        """

        :param state:
        :return:
        """
        potential_issues = self._analyze_state(state)

        annotation = get_potential_issues_annotation(state)
        annotation.potential_issues.extend(potential_issues)

    def _analyze_state(self, state):
        """

        :param state:
        :return:
        """

        write_slot = state.mstate.stack[-1]
        constraints = state.world_state.constraints + [
            write_slot == symbol_factory.BitVecVal(324345425435, 256)
        ]

        potential_issue = PotentialIssue(
            contract=state.environment.active_account.contract_name,
            function_name=state.environment.active_function_name,
            address=state.get_current_instruction()["address"],
            swc_id=WRITE_TO_ARBITRARY_STORAGE,
            title="Write to an arbitrary storage location",
            severity="High",
            bytecode=state.environment.code.bytecode,
            description_head="The caller can write to arbitrary storage locations.",
            description_tail="It is possible to write to arbitrary storage locations. By modifying the values of "
            "storage variables, attackers may bypass security controls or manipulate the business logic of "
            "the smart contract.",
            detector=self,
            constraints=constraints,
        )
        return [potential_issue]


detector = ArbitraryStorage()



"""This module contians the transaction models used throughout LASER's symbolic
execution."""

from copy import deepcopy
from z3 import ExprRef
from typing import Union, Optional
from mythril.support.support_utils import Singleton
from mythril.laser.ethereum.state.calldata import ConcreteCalldata
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.calldata import BaseCalldata, SymbolicCalldata
from mythril.laser.ethereum.state.return_data import ReturnData
from mythril.laser.ethereum.state.environment import Environment
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.smt import symbol_factory, UGE, BitVec
import logging

log = logging.getLogger(__name__)


class TxIdManager(object, metaclass=Singleton):
    def __init__(self):
        self._next_transaction_id = 0

    def get_next_tx_id(self):
        self._next_transaction_id += 1
        return str(self._next_transaction_id)

    def restart_counter(self):
        self._next_transaction_id = 0

    def set_counter(self, tx_id):
        self._next_transaction_id = tx_id


tx_id_manager = TxIdManager()


class TransactionEndSignal(Exception):
    """Exception raised when a transaction is finalized."""

    def __init__(self, global_state: GlobalState, revert=False) -> None:
        self.global_state = global_state
        self.revert = revert


class TransactionStartSignal(Exception):
    """Exception raised when a new transaction is started."""

    def __init__(
        self,
        transaction: Union["MessageCallTransaction", "ContractCreationTransaction"],
        op_code: str,
        global_state: GlobalState,
    ) -> None:
        self.transaction = transaction
        self.op_code = op_code
        self.global_state = global_state


class BaseTransaction:
    """Basic transaction class holding common data."""

    def __init__(
        self,
        world_state: WorldState,
        callee_account: Account = None,
        caller: ExprRef = None,
        call_data=None,
        identifier: Optional[str] = None,
        gas_price=None,
        gas_limit=None,
        origin=None,
        code=None,
        call_value=None,
        init_call_data=True,
        static=False,
        base_fee=None,
    ) -> None:
        assert isinstance(world_state, WorldState)
        self.world_state = world_state
        self.id = identifier or tx_id_manager.get_next_tx_id()

        self.gas_price = (
            gas_price
            if gas_price is not None
            else symbol_factory.BitVecSym(f"gasprice{identifier}", 256)
        )

        self.base_fee = (
            base_fee
            if base_fee is not None
            else symbol_factory.BitVecSym(f"basefee{identifier}", 256)
        )

        self.gas_limit = gas_limit

        self.origin = (
            origin
            if origin is not None
            else symbol_factory.BitVecSym(f"origin{identifier}", 256)
        )
        self.code = code

        self.caller = caller
        self.callee_account = callee_account
        if call_data is None and init_call_data:
            self.call_data = SymbolicCalldata(self.id)  # type: BaseCalldata
        else:
            self.call_data = (
                call_data
                if isinstance(call_data, BaseCalldata)
                else ConcreteCalldata(self.id, [])
            )

        self.call_value = (
            call_value
            if call_value is not None
            else symbol_factory.BitVecSym(f"callvalue{identifier}", 256)
        )
        self.static = static
        self.return_data = None  # type: str

    def initial_global_state_from_environment(self, environment, active_function):
        """

        :param environment:
        :param active_function:
        :return:
        """
        # Initialize the execution environment
        global_state = GlobalState(self.world_state, environment, None)
        global_state.environment.active_function_name = active_function

        sender = environment.sender
        receiver = environment.active_account.address
        value = (
            environment.callvalue
            if isinstance(environment.callvalue, BitVec)
            else symbol_factory.BitVecVal(environment.callvalue, 256)
        )

        global_state.world_state.constraints.append(
            UGE(global_state.world_state.balances[sender], value)
        )
        global_state.world_state.balances[receiver] += value
        global_state.world_state.balances[sender] -= value

        return global_state

    def initial_global_state(self) -> GlobalState:
        raise NotImplementedError

    def __str__(self) -> str:
        if self.callee_account is None or self.callee_account.address.symbolic is False:
            return "{} {} from {} to {:#42x}".format(
                self.__class__.__name__,
                self.id,
                self.caller,
                int(str(self.callee_account.address)) if self.callee_account else -1,
            )
        else:
            return "{} {} from {} to {}".format(
                self.__class__.__name__,
                self.id,
                self.caller,
                str(self.callee_account.address),
            )


class MessageCallTransaction(BaseTransaction):
    """Transaction object models an transaction."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def initial_global_state(self) -> GlobalState:
        """Initialize the execution environment."""
        environment = Environment(
            self.callee_account,
            self.caller,
            self.call_data,
            self.gas_price,
            self.call_value,
            self.origin,
            self.base_fee,
            code=self.code or self.callee_account.code,
            static=self.static,
        )
        return super().initial_global_state_from_environment(
            environment, active_function="fallback"
        )

    def end(self, global_state: GlobalState, return_data=None, revert=False) -> None:
        """

        :param global_state:
        :param return_data:
        :param revert:
        """
        self.return_data = return_data

        raise TransactionEndSignal(global_state, revert)


class ContractCreationTransaction(BaseTransaction):
    """Transaction object models an transaction."""

    def __init__(
        self,
        world_state: WorldState,
        caller: ExprRef = None,
        call_data=None,
        identifier: Optional[str] = None,
        gas_price=None,
        gas_limit=None,
        origin=None,
        code=None,
        call_value=None,
        contract_name=None,
        contract_address=None,
        base_fee=None,
    ) -> None:
        self.prev_world_state = deepcopy(world_state)
        contract_address = (
            contract_address if isinstance(contract_address, int) else None
        )
        callee_account = world_state.create_account(
            0, concrete_storage=True, creator=caller.value, address=contract_address
        )
        callee_account.contract_name = contract_name or callee_account.contract_name
        # init_call_data "should" be false, but it is easier to model the calldata symbolically
        # and add logic in codecopy/codesize/calldatacopy/calldatasize than to model code "correctly"
        super().__init__(
            world_state=world_state,
            callee_account=callee_account,
            caller=caller,
            call_data=call_data,
            identifier=identifier,
            gas_price=gas_price,
            gas_limit=gas_limit,
            origin=origin,
            code=code,
            call_value=call_value,
            init_call_data=True,
            base_fee=base_fee,
        )

    def initial_global_state(self) -> GlobalState:
        """Initialize the execution environment."""
        environment = Environment(
            active_account=self.callee_account,
            sender=self.caller,
            calldata=self.call_data,
            gasprice=self.gas_price,
            callvalue=self.call_value,
            origin=self.origin,
            basefee=self.base_fee,
            code=self.code,
        )
        return super().initial_global_state_from_environment(
            environment, active_function="constructor"
        )

    def end(self, global_state: GlobalState, return_data=None, revert=False):
        """

        :param global_state:
        :param return_data:
        :param revert:
        """

        if return_data is None or return_data.size == 0:
            self.return_data = None
            raise TransactionEndSignal(global_state, revert=revert)

        global_state.environment.active_account.code.assign_bytecode(
            tuple(return_data.return_data)
        )
        return_data = str(hex(global_state.environment.active_account.address.value))
        self.return_data = ReturnData(return_data, len(return_data) // 2)
        assert global_state.environment.active_account.code.instruction_list != []

        raise TransactionEndSignal(global_state, revert=revert)



"""This module contains the detection code for potentially insecure low-level
calls."""

from mythril.analysis import solver
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.potential_issues import Issue
from mythril.analysis.swc_data import ASSERT_VIOLATION
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import Extract, And
from mythril.exceptions import UnsatError
import logging
import eth_abi

log = logging.getLogger(__name__)

DESCRIPTION = """

Search for reachable user-supplied exceptions.
Report a warning if an log message is emitted: 'emit AssertionFailed(string)'

"""

assertion_failed_hash = (
    0xB42604CB105A16C8F6DB8A41E6B00C0C1B4826465E8BC504B3EB3E88B3E6A4A0
)

mstore_pattern = "0xcafecafecafecafecafecafecafecafecafecafecafecafecafecafecafe"


class UserAssertions(DetectionModule):
    """This module searches for user supplied exceptions: emit AssertionFailed("Error")."""

    name = "A user-defined assertion has been triggered"
    swc_id = ASSERT_VIOLATION
    description = DESCRIPTION
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["LOG1", "MSTORE"]

    def _execute(self, state: GlobalState) -> None:
        """

        :param state:
        :return:
        """

        return self._analyze_state(state)

    def _analyze_state(self, state: GlobalState):
        """

        :param state:
        :return:
        """
        opcode = state.get_current_instruction()["opcode"]
        message = None
        if opcode == "MSTORE":
            value = state.mstate.stack[-2]
            if value.symbolic:
                return []
            if mstore_pattern not in hex(value.value)[:126]:
                return []
            message = "Failed property id {}".format(Extract(15, 0, value).value)

        else:
            topic, size, mem_start = state.mstate.stack[-3:]

            if topic.symbolic or topic.value != assertion_failed_hash:
                return []
            if not mem_start.symbolic and not size.symbolic:
                try:
                    message = eth_abi.decode_single(
                        "string",
                        bytes(
                            state.mstate.memory[
                                mem_start.value + 32 : mem_start.value + size.value
                            ]
                        ),
                    )
                except:
                    pass
        try:
            transaction_sequence = solver.get_transaction_sequence(
                state, state.world_state.constraints
            )

            if message:
                description_tail = (
                    "A user-provided assertion failed with the message '{}'".format(
                        message
                    )
                )

            else:
                description_tail = "A user-provided assertion failed."

            log.debug("MythX assertion emitted: {}".format(description_tail))

            address = state.get_current_instruction()["address"]

            issue = Issue(
                contract=state.environment.active_account.contract_name,
                function_name=state.environment.active_function_name,
                address=address,
                swc_id=ASSERT_VIOLATION,
                title="Exception State",
                severity="Medium",
                description_head="A user-provided assertion failed.",
                description_tail=description_tail,
                bytecode=state.environment.code.bytecode,
                transaction_sequence=transaction_sequence,
                gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
            )
            state.annotate(
                IssueAnnotation(
                    detector=self,
                    issue=issue,
                    conditions=[And(*state.world_state.constraints)],
                )
            )
            return [issue]

        except UnsatError:
            log.debug("no model found")

        return []


detector = UserAssertions()



"""This module provides classes for an SMT abstraction of boolean
expressions."""

from typing import Union, cast, Set

import z3

from mythril.laser.smt.expression import Expression

# fmt: off


class Bool(Expression[z3.BoolRef]):
    """This is a Bool expression."""

    @property
    def is_false(self) -> bool:
        """Specifies whether this variable can be simplified to false.

        :return:
        """
        self.simplify()
        return z3.is_false(self.raw)

    @property
    def is_true(self) -> bool:
        """Specifies whether this variable can be simplified to true.

        :return:
        """
        self.simplify()
        return z3.is_true(self.raw)

    @property
    def value(self) -> Union[bool, None]:
        """Returns the concrete value of this bool if concrete, otherwise None.

        :return: Concrete value or None
        """
        self.simplify()
        if self.is_true:
            return True
        elif self.is_false:
            return False
        else:
            return None

    # MYPY: complains about overloading __eq__ # noqa
    def __eq__(self, other: object) -> "Bool":  # type: ignore
        """

        :param other:
        :return:
        """
        if isinstance(other, Expression):
            return Bool(cast(z3.BoolRef, self.raw == other.raw),
                        self.annotations.union(other.annotations))
        return Bool(cast(z3.BoolRef, self.raw == other), self.annotations)

    # MYPY: complains about overloading __ne__ # noqa
    def __ne__(self, other: object) -> "Bool":  # type: ignore
        """

        :param other:
        :return:
        """
        if isinstance(other, Expression):
            return Bool(cast(z3.BoolRef, self.raw != other.raw),
                        self.annotations.union(other.annotations))
        return Bool(cast(z3.BoolRef, self.raw != other), self.annotations)

    def __bool__(self) -> bool:
        """

        :return:
        """
        if self.value is not None:
            return self.value
        else:
            return False

    def substitute(self, original_expression, new_expression):
        """

        :param original_expression:
        :param new_expression:
        """
        if self.raw is None:
            return
        original_z3 = original_expression.raw
        new_z3 = new_expression.raw
        self.raw = z3.substitute(self.raw, (original_z3, new_z3))

    def __hash__(self) -> int:
        return self.raw.__hash__()


def And(*args: Union[Bool, bool]) -> Bool:
    """Create an And expression."""
    annotations = set()   # type: Set
    args_list = [arg if isinstance(arg, Bool) else Bool(arg) for arg in args]
    for arg in args_list:
        annotations = annotations.union(arg.annotations)
    return Bool(z3.And([a.raw for a in args_list]), annotations)


def Xor(a: Bool, b: Bool) -> Bool:
    """Create an And expression."""

    union = a.annotations.union(b.annotations)
    return Bool(z3.Xor(a.raw, b.raw), union)


def Or(*args: Union[Bool, bool]) -> Bool:
    """Create an or expression.

    :param a:
    :param b:
    :return:
    """
    args_list = [arg if isinstance(arg, Bool) else Bool(arg) for arg in args]
    annotations = set()  # type: Set
    for arg in args_list:
        annotations = annotations.union(arg.annotations)
    return Bool(z3.Or([a.raw for a in args_list]), annotations=annotations)


def Not(a: Bool) -> Bool:
    """Create a Not expression.

    :param a:
    :return:
    """
    return Bool(z3.Not(a.raw), a.annotations)


def is_false(a: Bool) -> bool:
    """Returns whether the provided bool can be simplified to false.

    :param a:
    :return:
    """
    return z3.is_false(a.raw)


def is_true(a: Bool) -> bool:
    """Returns whether the provided bool can be simplified to true.

    :param a:
    :return:
    """
    return z3.is_true(a.raw)



import z3

from mythril.laser.smt.solver.solver import Solver, Optimize, BaseSolver
from mythril.laser.smt.solver.independence_solver import IndependenceSolver
from mythril.laser.smt.solver.solver_statistics import SolverStatistics
from mythril.support.support_args import args

if args.parallel_solving:
    z3.set_param("parallel.enable", True)



import logging
from typing import Tuple


from mythril.laser.smt import And, BitVec, Bool, Function, URem, symbol_factory

log = logging.getLogger(__name__)


class ExponentFunctionManager:
    """
    Uses an uninterpreted function for exponentiation with the following properties:
    1) power(a, b) > 0
    2) if a = 256 => forall i if b = i then power(a, b) = (256 ^ i) % (2^256)

    Only these two properties are added as to handle indexing of boolean arrays.
    Caution should be exercised when increasing the conditions since it severely affects
    the solving time.
    """

    def __init__(self):
        power = Function("Power", [256, 256], 256)
        NUMBER_256 = symbol_factory.BitVecVal(256, 256)
        self.concrete_constraints = And(
            *[
                power(NUMBER_256, symbol_factory.BitVecVal(i, 256))
                == symbol_factory.BitVecVal(256**i, 256)
                for i in range(0, 32)
            ]
        )

    def create_condition(self, base: BitVec, exponent: BitVec) -> Tuple[BitVec, Bool]:
        """
        Creates a condition for exponentiation
        :param base: The base of exponentiation
        :param exponent: The exponent of the exponentiation
        :return: Tuple of condition and the exponentiation result
        """
        power = Function("Power", [256, 256], 256)
        exponentiation = power(base, exponent)

        if exponent.symbolic is False and base.symbolic is False:
            const_exponentiation = symbol_factory.BitVecVal(
                pow(base.value, exponent.value, 2**256),
                256,
                annotations=base.annotations.union(exponent.annotations),
            )
            constraint = const_exponentiation == exponentiation
            return const_exponentiation, constraint

        constraint = exponentiation > 0
        constraint = And(constraint, self.concrete_constraints)
        if base.value == 256:
            constraint = And(
                constraint,
                power(base, URem(exponent, symbol_factory.BitVecVal(32, 256)))
                == power(base, exponent),
            )

        return exponentiation, constraint


exponent_function_manager = ExponentFunctionManager()



"""This module contains utility functions for the Mythril support package."""

from collections import OrderedDict
from copy import deepcopy
from eth_hash.auto import keccak
from functools import lru_cache
from typing import Dict
from z3 import is_true, simplify, And
import logging

log = logging.getLogger(__name__)


class Singleton(type):
    """A metaclass type implementing the singleton pattern."""

    _instances = {}  # type: Dict

    def __call__(cls, *args, **kwargs):
        """Delegate the call to an existing resource or a a new one.

        This is not thread- or process-safe by default. It must be protected with
        a lock.

        :param args:
        :param kwargs:
        :return:
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class LRUCache:
    def __init__(self, size):
        self.size = size
        self.lru_cache = OrderedDict()

    def get(self, key):
        try:
            value = self.lru_cache.pop(key)
            self.lru_cache[key] = value
            return value
        except KeyError:
            return -1

    def put(self, key, value):
        try:
            self.lru_cache.pop(key)
        except KeyError:
            if len(self.lru_cache) >= self.size:
                self.lru_cache.popitem(last=False)
        self.lru_cache[key] = value


class ModelCache:
    def __init__(self):
        self.model_cache = LRUCache(size=100)

    @lru_cache(maxsize=2**10)
    def check_quick_sat(self, constraints) -> bool:
        model_list = list(reversed(self.model_cache.lru_cache.keys()))
        for model in reversed(self.model_cache.lru_cache.keys()):
            model_copy = deepcopy(model)
            if is_true(model_copy.eval(constraints, model_completion=True)):
                self.model_cache.put(model, self.model_cache.get(model) + 1)
                return model
        return False

    def put(self, key, value):
        self.model_cache.put(key, value)


@lru_cache(maxsize=2**10)
def get_code_hash(code) -> str:
    """
    :param code: bytecode
    :return: Returns hash of the given bytecode
    """
    if type(code) == tuple:
        # Temporary hack, since we cannot find symbols of sha3
        return str(hash(code))

    code = code[2:] if code[:2] == "0x" else code
    try:
        hash_ = keccak(bytes.fromhex(code))
        return "0x" + hash_.hex()
    except ValueError:
        log.debug("Unable to change the bytecode to bytes. Bytecode: {}".format(code))
        return ""


def sha3(value):
    if type(value) == str:
        if value[:2] == "0x":
            new_hash = keccak(bytes.fromhex(value))
        else:
            new_hash = keccak(value.encode())
    else:
        new_hash = keccak(value)
    return new_hash


def zpad(x, l):
    """
    Left zero pad value `x` at least to length `l`.
    """
    return b"\x00" * max(0, l - len(x)) + x


def rzpad(value, total_length):
    """
    Right zero pad value `x` at least to length `l`.
    """
    return value + b"\x00" * max(0, total_length - len(value))



from mythril.mythril import MythrilDisassembler
from mythril.laser.ethereum.transaction import execute_contract_creation
from mythril.ethereum import util
import mythril.laser.ethereum.svm as svm
from mythril.disassembler.disassembly import Disassembly
from datetime import datetime
from mythril.solidity.soliditycontract import SolidityContract
import tests
from mythril.analysis.security import fire_lasers
from mythril.analysis.symbolic import SymExecWrapper

solc_binary = MythrilDisassembler._init_solc_binary("v0.5.0")


def test_create():
    contract = SolidityContract(
        str(tests.TESTDATA_INPUTS_CONTRACTS / "calls.sol"), solc_binary=solc_binary
    )

    laser_evm = svm.LaserEVM({})

    laser_evm.time = datetime.now()
    execute_contract_creation(laser_evm, contract.creation_code)

    resulting_final_state = laser_evm.open_states[0]

    for address, created_account in resulting_final_state.accounts.items():
        created_account_code = created_account.code
        actual_code = Disassembly(contract.code)

        for i in range(len(created_account_code.instruction_list)):
            found_instruction = created_account_code.instruction_list[i]
            actual_instruction = actual_code.instruction_list[i]

            assert found_instruction["opcode"] == actual_instruction["opcode"]


def test_sym_exec():
    contract = SolidityContract(
        str(tests.TESTDATA_INPUTS_CONTRACTS / "calls.sol"), solc_binary=solc_binary
    )

    sym = SymExecWrapper(
        contract,
        address=(util.get_indexed_address(0)),
        strategy="dfs",
        execution_timeout=25,
    )
    issues = fire_lasers(sym)

    assert len(issues) != 0



import pytest
from eth._utils.numeric import ceil32
from mythril.laser.smt import simplify, symbol_factory, Concat, Extract

from mythril.laser.ethereum.state.machine_state import MachineState
from mythril.laser.ethereum.evm_exceptions import StackUnderflowException
from mythril.laser.ethereum.state.memory import Memory

memory_extension_test_data = [(0, 0, 10), (0, 30, 10), (100, 22, 8)]


@pytest.mark.parametrize(
    "initial_size,start,extension_size", memory_extension_test_data
)
def test_memory_extension(initial_size, start, extension_size):
    # Arrange
    machine_state = MachineState(gas_limit=8000000)
    machine_state.memory = Memory()
    machine_state.memory.extend(initial_size)

    # Act
    machine_state.mem_extend(start, extension_size)

    # Assert
    assert machine_state.memory_size == len(machine_state.memory)
    assert machine_state.memory_size == max(
        initial_size, (ceil32(start + extension_size) // 32) * 32
    )


stack_pop_too_many_test_data = [(0, 1), (0, 2), (5, 1), (5, 10)]


@pytest.mark.parametrize("initial_size,overflow", stack_pop_too_many_test_data)
def test_stack_pop_too_many(initial_size, overflow):
    # Arrange
    machine_state = MachineState(8000000)
    machine_state.stack = [42] * initial_size

    # Act + Assert
    with pytest.raises(StackUnderflowException):
        machine_state.pop(initial_size + overflow)


stack_pop_test_data = [
    ([1, 2, 3], 2, [3, 2]),
    ([1, 3, 4, 7, 7, 1, 2], 5, [2, 1, 7, 7, 4]),
]


@pytest.mark.parametrize("initial_stack,amount,expected", stack_pop_test_data)
def test_stack_multiple_pop(initial_stack, amount, expected):
    # Arrange
    machine_state = MachineState(8000000)
    machine_state.stack = initial_stack[:]

    # Act
    results = machine_state.pop(amount)

    # Assert
    assert results == initial_stack[-amount:][::-1]
    assert results == expected
    assert len(machine_state.stack) == len(initial_stack) - amount


def test_stack_multiple_pop_():
    # Arrange
    machine_state = MachineState(8000000)
    machine_state.stack = [1, 2, 3]

    # Act
    a, b = machine_state.pop(2)

    # Assert
    assert a == 3
    assert b == 2


def test_stack_single_pop():
    # Arrange
    machine_state = MachineState(8000000)
    machine_state.stack = [1, 2, 3]

    # Act
    result = machine_state.pop()

    # Assert
    assert isinstance(result, int)


def test_memory_zeroed():
    # Arrange
    mem = Memory()
    mem.extend(2000 + 32)

    # Act
    mem[11] = 10
    mem.write_word_at(2000, 0x12345)

    # Assert
    assert mem[10] == 0
    assert mem[100] == 0
    assert mem.get_word_at(1000) == 0


def test_memory_write():
    # Arrange
    mem = Memory()
    mem.extend(200 + 32)

    a = symbol_factory.BitVecSym("a", 256)
    b = symbol_factory.BitVecSym("b", 8)

    # Act
    mem[11] = 10
    mem[12] = b
    mem.write_word_at(200, 0x12345)
    mem.write_word_at(100, a)

    # Assert
    assert mem[0] == 0
    assert mem[11] == 10
    assert mem[200 + 31] == 0x45
    assert mem.get_word_at(200) == 0x12345
    assert simplify(a == mem.get_word_at(100))
    assert simplify(b == mem[12])


def test_memory_symbolic():
    # Arrange
    mem = Memory()
    mem.extend(200 + 32)

    a = symbol_factory.BitVecSym("a", 256)
    b = symbol_factory.BitVecSym("b", 256)

    # Act
    mem.write_word_at(a + symbol_factory.BitVecVal(1, 256), b)

    # Assert
    assert mem.get_word_at(a) == Concat(
        symbol_factory.BitVecVal(0, 256), Extract(255, 8, b)
    )
    assert mem[a] == 0



TRANSFER_METHODS = ["transfer", "send"]


class SolidityFeatureExtractor:
    def __init__(self, ast):
        self.ast = ast

    def extract_features(self):
        function_features = {}
        function_nodes = self.find_function_nodes(self.ast)
        modifier_vars = {}
        for modifier_node in self.find_modifier_nodes(self.ast):
            modifier_vars[modifier_node["name"]] = self.find_variables_in_require(
                modifier_node
            )
            modifier_vars[modifier_node["name"]].update(
                self.find_variables_in_if(modifier_node)
            )

        print(modifier_vars)

        for node in function_nodes:
            function_name = self.get_function_name(node)
            contains_selfdestruct = self.contains_selfdestruct(node)
            contains_call = self.contains_call(node)
            contains_delegatecall = self.contains_delegatecall(node)
            contains_callcode = self.contains_callcode(node)
            contains_staticcall = self.contains_staticcall(node)
            all_require_vars = self.find_variables_in_require(node)
            ether_vars = self.extract_address_variable(node)
            print(ether_vars)
            for modifier in node.get("modifiers", []):
                all_require_vars.update(modifier_vars[modifier["modifierName"]["name"]])
            is_payable = self.is_function_payable(node)
            has_isowner_modifier = self.has_isowner_modifier(node)
            contains_assert = self.contains_assert(node)
            function_features[function_name] = {
                "contains_selfdestruct": contains_selfdestruct,
                "contains_call": contains_call,
                "is_payable": is_payable,
                "has_owner_modifier": has_isowner_modifier,
                "contains_assert": contains_assert,
                "contains_callcode": contains_callcode,
                "contains_delegatecall": contains_delegatecall,
                "contains_staticcall": contains_staticcall,
                "all_require_vars": all_require_vars,
                "transfer_vars": ether_vars,
            }

        return function_features

    def find_function_nodes(self, node):
        if node["nodeType"] == "FunctionDefinition":
            yield node

        if "nodes" in node:
            for child_node in node["nodes"]:
                yield from self.find_function_nodes(child_node)

    def find_modifier_nodes(self, node):
        if node["nodeType"] == "ModifierDefinition":
            yield node

        if "nodes" in node:
            for child_node in node["nodes"]:
                yield from self.find_modifier_nodes(child_node)

    def get_function_name(self, node):
        return node["name"]

    def contains_command(self, node, command):
        if isinstance(node, dict):
            if command in node.values():
                return True

            for value in node.values():
                if isinstance(value, (dict, list)):
                    if self.contains_command(value, command):
                        return True

        elif isinstance(node, list):
            for item in node:
                if self.contains_command(item, command):
                    return True

        return False

    def contains_call(self, node):
        return self.contains_command(node, "call")

    def is_function_payable(self, node):
        return node.get("stateMutability") == "payable"

    def has_isowner_modifier(self, node):
        if "modifiers" in node:
            for modifier in node["modifiers"]:
                if modifier["modifierName"]["name"].lower() in ("isowner", "onlyowner"):
                    return True
        return False

    def contains_assert(self, node):
        return self.contains_command(node, "assert")

    def contains_selfdestruct(self, node):
        return self.contains_command(node, "selfdestruct")

    def contains_delegatecall(self, node):
        return self.contains_command(node, "delegatecall")

    def contains_callcode(self, node):
        return self.contains_command(node, "callcode")

    def contains_staticcall(self, node):
        return self.contains_command(node, "staticcall")

    def contains_require(self, node):
        return self.contains_command(node, "require")

    def extract_nodes(self, node, command, parent=None):
        node_list = []
        if isinstance(node, dict):
            if command in node.values():
                node_list.append((parent, node))

            for key, value in node.items():
                if isinstance(value, (dict, list)):
                    node_list.extend(self.extract_nodes(value, command, parent=node))
        elif isinstance(node, list):
            for item in node:
                node_list.extend(self.extract_nodes(item, command, parent=node))
        return node_list

    def find_all_variables(self, node):
        variables = set()

        def traverse(node):
            if isinstance(node, dict):
                for key, value in node.items():
                    if key == "nodeType" and value == "Identifier":
                        if "name" in node:
                            variables.add(node["name"])
                    elif isinstance(value, (dict, list)):
                        traverse(value)
            elif isinstance(node, list):
                for item in node:
                    traverse(item)

        traverse(node)
        return variables

    def find_variables_in_require(self, node):

        nodes = self.extract_nodes(node, "require")
        variables = set()
        for parent, _ in nodes:
            if "arguments" in parent:
                arguments = parent["arguments"]
                for argument in arguments:
                    variables.update(self.find_all_variables(argument))
        return variables

    def find_variables_in_if(self, node):
        variables = []

        def traverse(node):
            if "condition" in node:
                condition = node["condition"]
                if (
                    "leftExpression" in condition
                    and condition["leftExpression"]["nodeType"] == "Identifier"
                ):
                    variables.append(condition["leftExpression"]["name"])
                if (
                    "rightExpression" in condition
                    and condition["rightExpression"]["nodeType"] == "Identifier"
                ):
                    variables.append(condition["rightExpression"]["name"])

                traverse(condition)

            if "falseBody" in node and node["falseBody"]:
                traverse(node["falseBody"])

            if "trueBody" in node and node["trueBody"]:
                if (
                    "nodeType" in node["trueBody"]
                    and node["trueBody"]["nodeType"] == "Block"
                ):
                    statements = node["trueBody"].get("statements", [])
                    for statement in statements:
                        traverse(statement)
                else:
                    traverse(node["trueBody"])

            if "body" in node and node["body"]:
                if "nodeType" in node["body"] and node["body"]["nodeType"] == "Block":
                    statements = node["body"].get("statements", [])
                    for statement in statements:
                        traverse(statement)
                else:
                    traverse(node["body"])

        traverse(node)

        return variables

    def extract_address_variable(self, node):
        if type(node) == int:
            return set([])
        transfer_vars = set([])
        if (
            node.get("nodeType", "") == "ExpressionStatement"
            and node.get("expression", {}).get("nodeType") == "FunctionCall"
        ):
            expression = node["expression"].get("expression", None)
            if expression is not None:
                if (
                    expression["nodeType"] == "MemberAccess"
                    and expression["memberName"] in TRANSFER_METHODS
                ):
                    print(expression)
                    address_variable = expression["expression"].get("name")
                    if address_variable:
                        transfer_vars.update(set([address_variable]))

        for key, value in node.items():
            if isinstance(value, dict):
                transfer_vars.update(self.extract_address_variable(value))

            elif isinstance(value, list):
                for item in value:
                    transfer_vars.update(self.extract_address_variable(item))

        return transfer_vars



"""This module provides a function to convert a state space into a set of state
nodes and transition edges."""
from z3 import Z3Exception
from mythril.laser.smt import simplify
from mythril.laser.ethereum.svm import NodeFlags
import re

colors = [
    {
        "border": "#26996f",
        "background": "#2f7e5b",
        "highlight": {"border": "#fff", "background": "#28a16f"},
    },
    {
        "border": "#9e42b3",
        "background": "#842899",
        "highlight": {"border": "#fff", "background": "#933da6"},
    },
    {
        "border": "#b82323",
        "background": "#991d1d",
        "highlight": {"border": "#fff", "background": "#a61f1f"},
    },
    {
        "border": "#4753bf",
        "background": "#3b46a1",
        "highlight": {"border": "#fff", "background": "#424db3"},
    },
    {
        "border": "#26996f",
        "background": "#2f7e5b",
        "highlight": {"border": "#fff", "background": "#28a16f"},
    },
    {
        "border": "#9e42b3",
        "background": "#842899",
        "highlight": {"border": "#fff", "background": "#933da6"},
    },
    {
        "border": "#b82323",
        "background": "#991d1d",
        "highlight": {"border": "#fff", "background": "#a61f1f"},
    },
    {
        "border": "#4753bf",
        "background": "#3b46a1",
        "highlight": {"border": "#fff", "background": "#424db3"},
    },
]


def get_serializable_statespace(statespace):
    """

    :param statespace:
    :return:
    """
    nodes = []
    edges = []

    color_map = {}
    i = 0
    for k in statespace.accounts:
        color_map[statespace.accounts[k].contract_name] = colors[i]
        i += 1

    for node_key in statespace.nodes:

        node = statespace.nodes[node_key]

        code = node.get_cfg_dict()["code"]
        code = re.sub("([0-9a-f]{8})[0-9a-f]+", lambda m: m.group(1) + "(...)", code)

        if NodeFlags.FUNC_ENTRY in node.flags:
            code = re.sub("JUMPDEST", node.function_name, code)

        code_split = code.split("\\n")

        truncated_code = (
            code
            if (len(code_split) < 7)
            else "\\n".join(code_split[:6]) + "\\n(click to expand +)"
        )
        try:
            color = color_map[node.get_cfg_dict()["contract_name"]]
        except KeyError:
            color = colors[i]
            i += 1
            color_map[node.get_cfg_dict()["contract_name"]] = color

        def get_state_accounts(node_state):
            """

            :param node_state:
            :return:
            """
            state_accounts = []
            for key in node_state.accounts:
                account = node_state.accounts[key].as_dict
                account.pop("code", None)
                account["balance"] = str(account["balance"])

                storage = {}
                for storage_key in account["storage"].printable_storage:
                    storage[str(storage_key)] = str(account["storage"][storage_key])

                state_accounts.append({"address": key, "storage": storage})
            return state_accounts

        states = [
            {"machine": x.mstate.as_dict, "accounts": get_state_accounts(x)}
            for x in node.states
        ]

        for state in states:
            state["machine"]["stack"] = [str(s) for s in state["machine"]["stack"]]
            state["machine"]["memory"] = [
                str(m)
                for m in state["machine"]["memory"][: len(state["machine"]["memory"])]
            ]

        truncated_code = truncated_code.replace("\\n", "\n")
        code = code.replace("\\n", "\n")

        s_node = {
            "id": str(node_key),
            "func": str(node.function_name),
            "label": truncated_code,
            "code": code,
            "truncated": truncated_code,
            "states": states,
            "color": color,
            "instructions": code.split("\n"),
        }

        nodes.append(s_node)

    for edge in statespace.edges:

        if edge.condition is None:
            label = ""
        else:

            try:
                label = str(simplify(edge.condition)).replace("\n", "")
            except Z3Exception:
                label = str(edge.condition).replace("\n", "")

        label = re.sub(
            "([^_])([\d]{2}\d+)", lambda m: m.group(1) + hex(int(m.group(2))), label
        )

        s_edge = {
            "from": str(edge.as_dict["from"]),
            "to": str(edge.as_dict["to"]),
            "arrows": "to",
            "label": label,
            "smooth": {"type": "cubicBezier"},
        }

        edges.append(s_edge)

    return {"edges": edges, "nodes": nodes}



"""This module contains the detection code for insecure delegate call usage."""
import logging
from typing import List

from mythril.analysis.potential_issues import (
    get_potential_issues_annotation,
    PotentialIssue,
)
from mythril.analysis.swc_data import DELEGATECALL_TO_UNTRUSTED_CONTRACT
from mythril.laser.ethereum.transaction.symbolic import ACTORS
from mythril.laser.ethereum.transaction.transaction_models import (
    ContractCreationTransaction,
)
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.exceptions import UnsatError
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import symbol_factory, UGT, Bool

log = logging.getLogger(__name__)


class ArbitraryDelegateCall(DetectionModule):
    """This module detects delegatecall to a user-supplied address."""

    name = "Delegatecall to a user-specified address"
    swc_id = DELEGATECALL_TO_UNTRUSTED_CONTRACT
    description = "Check for invocations of delegatecall to a user-supplied address."
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["DELEGATECALL"]

    def _execute(self, state: GlobalState) -> None:
        """

        :param state:
        :return:
        """
        potential_issues = self._analyze_state(state)

        annotation = get_potential_issues_annotation(state)
        annotation.potential_issues.extend(potential_issues)

    def _analyze_state(self, state: GlobalState) -> List[PotentialIssue]:
        """
        :param state: the current state
        :return: returns the issues for that corresponding state
        """
        gas = state.mstate.stack[-1]
        to = state.mstate.stack[-2]

        constraints = [
            to == ACTORS.attacker,
            UGT(gas, symbol_factory.BitVecVal(2300, 256)),
            state.new_bitvec(
                "retval_{}".format(state.get_current_instruction()["address"]), 256
            )
            == 1,
        ]

        for tx in state.world_state.transaction_sequence:
            if not isinstance(tx, ContractCreationTransaction):
                constraints.append(tx.caller == ACTORS.attacker)

        try:
            address = state.get_current_instruction()["address"]

            logging.debug(
                "[DELEGATECALL] Detected potential delegatecall to a user-supplied address : {}".format(
                    address
                )
            )

            description_head = "The contract delegates execution to another contract with a user-supplied address."
            description_tail = (
                "The smart contract delegates execution to a user-supplied address.This could allow an attacker to "
                "execute arbitrary code in the context of this contract account and manipulate the state of the "
                "contract account or execute actions on its behalf."
            )

            return [
                PotentialIssue(
                    contract=state.environment.active_account.contract_name,
                    function_name=state.environment.active_function_name,
                    address=address,
                    swc_id=DELEGATECALL_TO_UNTRUSTED_CONTRACT,
                    bytecode=state.environment.code.bytecode,
                    title="Delegatecall to user-supplied address",
                    severity="High",
                    description_head=description_head,
                    description_tail=description_tail,
                    constraints=constraints,
                    detector=self,
                )
            ]

        except UnsatError:
            return []


detector = ArbitraryDelegateCall()



""" Mythril Detection Modules

This module includes an definition of the DetectionModule interface.
DetectionModules implement different analysis rules to find weaknesses and vulnerabilities.
"""
import logging
from typing import List, Set, Optional, Tuple

from mythril.analysis.report import Issue
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.support.support_args import args
from mythril.support.support_utils import get_code_hash
from abc import ABC, abstractmethod
from enum import Enum

# Get logger instance
log = logging.getLogger(__name__)


class EntryPoint(Enum):
    """EntryPoint Enum

    This enum is used to signify the entry_point of detection modules.
    See also the class documentation of DetectionModule
    """

    POST = 1
    CALLBACK = 2


class DetectionModule(ABC):
    """The base detection module.

    All custom-built detection modules must inherit from this class.

    There are several class properties that expose information about the detection modules

    :param name: The name of the detection module
    :param swc_id: The SWC ID associated with the weakness that the module detects
    :param description: A description of the detection module, and what it detects
    :param entry_point: Mythril can run callback style detection modules, or modules that search the statespace.
                [IMPORTANT] POST entry points severely slow down the analysis, try to always use callback style modules
    :param pre_hooks: A list of instructions to hook the laser vm for (pre execution of the instruction)
    :param post_hooks: A list of instructions to hook the laser vm for (post execution of the instruction)
    """

    name = "Detection Module Name / Title"
    swc_id = "SWC-000"
    description = "Detection module description"
    entry_point: EntryPoint = EntryPoint.CALLBACK
    pre_hooks: List[str] = []
    post_hooks: List[str] = []

    def __init__(self) -> None:
        self.issues: List[Issue] = []
        self.cache: Set[Tuple[int, str]] = set()
        self.auto_cache = True

    def reset_module(self):
        """Resets the storage of this module"""
        self.issues = []

    def update_cache(self, issues=None):
        """
        Updates cache with param issues, updates against self.issues, if the param is None
        :param issues: The issues used to update the cache
        """
        issues = issues or self.issues
        for issue in issues:
            self.cache.add((issue.address, issue.bytecode_hash))

    def execute(self, target: GlobalState) -> Optional[List[Issue]]:
        """The entry point for execution, which is being called by Mythril.

        :param target: The target of the analysis, either a global state (callback) or the entire statespace (post)
        :return: List of encountered issues
        """

        log.debug("Entering analysis module: {}".format(self.__class__.__name__))

        if (
            target.get_current_instruction()["address"],
            get_code_hash(target.environment.code.bytecode),
        ) in self.cache and self.auto_cache:
            log.debug(
                f"Issue in cache for the analysis module: {self.__class__.__name__}, address: {target.get_current_instruction()['address']}"
            )
            return []

        result = self._execute(target)
        log.debug("Exiting analysis module: {}".format(self.__class__.__name__))

        if result and not args.use_issue_annotations:
            if self.auto_cache:
                self.update_cache(result)
            self.issues += result

        return result

    @abstractmethod
    def _execute(self, target) -> Optional[List[Issue]]:
        """Module main method (override this)

        :param target: The target of the analysis, either a global state (callback) or the entire statespace (post)
        :return: List of encountered issues
        """
        pass

    def __repr__(self) -> str:
        return (
            "<"
            "DetectionModule "
            "name={0.name} "
            "swc_id={0.swc_id} "
            "pre_hooks={0.pre_hooks} "
            "post_hooks={0.post_hooks} "
            "description={0.description}"
            ">"
        ).format(self)



import codecs
import logging
import os
import re

from pathlib import Path
from shutil import copyfile
from configparser import ConfigParser
from typing import Optional

from mythril.exceptions import CriticalError
from mythril.ethereum.interface.rpc.client import EthJsonRpc
from mythril.support.lock import LockFile

log = logging.getLogger(__name__)


class MythrilConfig:
    """
    The Mythril Analyzer class
    Responsible for setup of the mythril environment
    """

    def __init__(self):
        self.infura_id = os.getenv("INFURA_ID")  # type: str
        self.mythril_dir = self.init_mythril_dir()
        self.config_path = os.path.join(self.mythril_dir, "config.ini")
        self._init_config()
        self.eth = None  # type: Optional[EthJsonRpc]

    def set_api_infura_id(self, id):
        self.infura_id = id

    @staticmethod
    def init_mythril_dir() -> str:
        """
        Initializes the mythril dir and config.ini file
        :return: The mythril dir's path
        """

        try:
            mythril_dir = os.environ["MYTHRIL_DIR"]
        except KeyError:
            mythril_dir = os.path.join(os.path.expanduser("~"), ".mythril")

        if not os.path.exists(mythril_dir):
            # Initialize data directory
            log.info("Creating mythril data directory")
            os.mkdir(mythril_dir)

        db_path = str(Path(mythril_dir) / "signatures.db")
        if not os.path.exists(db_path):
            # if the default mythril dir doesn't contain a signature DB
            # initialize it with the default one from the project root
            asset_dir = Path(__file__).parent.parent / "support" / "assets"
            copyfile(str(asset_dir / "signatures.db"), db_path)

        return mythril_dir

    def _init_config(self):
        """If no config file exists, create it and add default options.
        Defaults:-
            - dynamic loading is set to infura by default in the file
        """

        if not os.path.exists(self.config_path):
            log.info("No config file found. Creating default: " + self.config_path)
            open(self.config_path, "a").close()

        config = ConfigParser(allow_no_value=True)

        config.optionxform = str
        with LockFile(self.config_path):
            config.read(self.config_path, "utf-8")

            if "defaults" not in config.sections():
                self._add_default_options(config)

            if not config.has_option("defaults", "dynamic_loading"):
                self._add_dynamic_loading_option(config)

            if not config.has_option("defaults", "infura_id"):
                config.set("defaults", "infura_id", "")

            with codecs.open(self.config_path, "w", "utf-8") as fp:
                config.write(fp)

            if not self.infura_id:
                self.infura_id = config.get("defaults", "infura_id", fallback="")

    @staticmethod
    def _add_default_options(config: ConfigParser) -> None:
        """
        Adds defaults option to config.ini
        :param config: The config file object
        :return: None
        """
        config.add_section("defaults")

    @staticmethod
    def _add_dynamic_loading_option(config: ConfigParser) -> None:
        """
        Sets the dynamic loading config option in .mythril/config.ini file
        :param config: The config file object
        :return: None
        """
        config.set(
            "defaults", "#– To connect to Infura use dynamic_loading: infura", ""
        )
        config.set(
            "defaults",
            "#– To connect to Rpc use "
            "dynamic_loading: HOST:PORT / ganache / infura-[network_name]",
            "",
        )
        config.set(
            "defaults", "#– To connect to local host use dynamic_loading: localhost", ""
        )
        config.set("defaults", "dynamic_loading", "infura")

    def set_api_rpc_infura(self) -> None:
        """Set the RPC mode to INFURA on Mainnet."""
        log.info("Using INFURA Main Net for RPC queries")
        self.eth = EthJsonRpc(
            "mainnet.infura.io/v3/{}".format(self.infura_id), None, True
        )

    def set_api_rpc(self, rpc: str = None, rpctls: bool = False) -> None:
        """
        Sets the RPC mode to either of ganache or infura
        :param rpc: either of the strings - ganache, infura-mainnet, infura-rinkeby, infura-kovan, infura-ropsten,
        infura-goerli, avalanche, arbitrum, bsc, optimism, polygon
        """

        if rpc == "ganache":
            rpcconfig = ("localhost", 7545, False)

        elif rpc.startswith("infura-"):
            network = rpc.replace("infura-", "")
            layer_one = ["mainnet", "rinkeby", "kovan", "ropsten", "goerli", "sepolia"]
            layer_two = [
                "avalanche",
                "arbitrum",
                "bsc",
                "optimism",
                "polygon",
                "celo",
                "starknet",
                "aurora",
                "near",
                "palm",
            ]
            if network in layer_one + layer_two:
                if self.infura_id in (None, ""):
                    log.info(
                        "Infura key not provided, so onchain access is disabled. "
                        "Use --infura-id <INFURA_ID> "
                        "or set it in the environment variable INFURA_ID "
                        "or in the ~/.mythril/config.ini file"
                    )
                    self.eth = None
                    return
                if network in layer_one:
                    rpcconfig = (
                        f"{network}.infura.io/v3/{self.infura_id}",
                        None,
                        True,
                    )
                else:
                    rpcconfig = (
                        f"{network}-mainnet.infura.io/v3/{self.infura_id}",
                        None,
                        True,
                    )
            else:
                raise CriticalError(
                    f"Invalid network {network}, use 'mainnet', 'rinkeby', 'kovan', 'ropsten', 'goerli', 'avalanche', 'arbitrum', 'optimism', or 'polygon'"
                )
        else:
            try:
                host, port = rpc.split(":")
                rpcconfig = (host, int(port), rpctls)
            except ValueError:
                raise CriticalError(
                    "Invalid RPC argument, use 'ganache', 'infura-[network]', or 'HOST:PORT'"
                )

        if rpcconfig:
            log.info("Using RPC settings: %s" % str(rpcconfig))
            self.eth = EthJsonRpc(rpcconfig[0], rpcconfig[1], rpcconfig[2])
        else:
            raise CriticalError("Invalid RPC settings, check help for details.")

    def set_api_rpc_localhost(self) -> None:
        """Set the RPC mode to a local instance."""
        log.info("Using default RPC settings: http://localhost:8545")
        self.eth = EthJsonRpc("localhost", 8545)

    def set_api_from_config_path(self) -> None:
        """Set the RPC mode based on a given config file."""
        config = ConfigParser(allow_no_value=False)
        # TODO: Remove this after this issue https://github.com/python/mypy/issues/2427 is closed
        config.optionxform = str  # type:ignore
        config.read(self.config_path, "utf-8")
        if config.has_option("defaults", "dynamic_loading"):
            dynamic_loading = config.get("defaults", "dynamic_loading")
        else:
            dynamic_loading = "infura"
        self._set_rpc(dynamic_loading)

    def _set_rpc(self, rpc_type: str) -> None:
        """
        Sets rpc based on the type
        :param rpc_type: The type of connection: like infura, ganache, localhost
        :return:
        """
        if rpc_type == "infura":
            self.set_api_rpc_infura()
        elif rpc_type == "localhost":
            self.set_api_rpc_localhost()
        else:
            self.set_api_rpc(rpc_type)



import json
import logging
import os
from pathlib import Path
import re
import shutil
import solc
import subprocess
import sys
import warnings

from eth_utils import int_to_big_endian
from semantic_version import Version, NpmSpec
from typing import List, Tuple, Optional, TYPE_CHECKING

from mythril.support.support_utils import sha3, zpad
from mythril.ethereum import util
from mythril.ethereum.interface.rpc.client import EthJsonRpc
from mythril.exceptions import CriticalError, CompilerError, NoContractFoundError
from mythril.support import signatures
from mythril.support.support_utils import rzpad
from mythril.support.support_args import args
from mythril.ethereum.evmcontract import EVMContract
from mythril.ethereum.interface.rpc.exceptions import ConnectionError
from mythril.solidity.soliditycontract import (
    SolidityContract,
    get_contracts_from_file,
    get_contracts_from_foundry,
)
from mythril.support.support_args import args


def format_Warning(message, category, filename, lineno, line=""):
    return "{}: {}\n\n".format(str(filename), str(message))


warnings.formatwarning = format_Warning


log = logging.getLogger(__name__)


class MythrilDisassembler:
    """
    The Mythril Disassembler class
    Responsible for generating disassembly of smart contracts:
        - Compiles solc code from file/onchain
        - Can also be used to access onchain storage data
    """

    def __init__(
        self,
        eth: Optional[EthJsonRpc] = None,
        solc_version: str = None,
        solc_settings_json: str = None,
        enable_online_lookup: bool = False,
        solc_args=None,
    ) -> None:
        args.solc_args = solc_args
        self.solc_binary = self._init_solc_binary(solc_version)
        self.solc_settings_json = solc_settings_json
        self.eth = eth
        self.enable_online_lookup = enable_online_lookup
        self.sigs = signatures.SignatureDB(enable_online_lookup=enable_online_lookup)
        self.contracts = []  # type: List[EVMContract]

    @staticmethod
    def _init_solc_binary(version: str) -> Optional[str]:
        """
        Only proper versions are supported. No nightlies, commits etc (such as available in remix).
        :param version: Version of the solc binary required
        :return: The solc binary of the corresponding version
        """

        if not version:
            return None

        # tried converting input to semver, seemed not necessary so just slicing for now
        try:
            main_version = solc.get_solc_version_string()
        except:
            main_version = ""  # allow missing solc will download instead
        main_version_number = re.search(r"\d+.\d+.\d+", main_version)

        if version.startswith("v"):
            version = version[1:]
        if version and NpmSpec("^0.8.0").match(Version(version)):
            args.use_integer_module = False
        if version == main_version_number:
            log.info("Given version matches installed version")
            solc_binary = os.environ.get("SOLC") or "solc"
        else:
            solc_binary = util.solc_exists(version)
            if solc_binary is None:
                raise CriticalError(
                    "The version of solc that is needed cannot be installed automatically"
                )
            else:
                log.info("Setting the compiler to %s", solc_binary)

        return solc_binary

    def load_from_bytecode(
        self, code: str, bin_runtime: bool = False, address: Optional[str] = None
    ) -> Tuple[str, EVMContract]:
        """
        Returns the address and the contract class for the given bytecode
        :param code: Bytecode
        :param bin_runtime: Whether the code is runtime code or creation code
        :param address: address of contract
        :return: tuple(address, Contract class)
        """
        if address is None:
            address = util.get_indexed_address(0)

        if bin_runtime:
            self.contracts.append(
                EVMContract(
                    code=code,
                    name="MAIN",
                    enable_online_lookup=self.enable_online_lookup,
                )
            )
        else:
            self.contracts.append(
                EVMContract(
                    creation_code=code,
                    name="MAIN",
                    enable_online_lookup=self.enable_online_lookup,
                )
            )
        return address, self.contracts[-1]  # return address and contract object

    def load_from_address(self, address: str) -> Tuple[str, EVMContract]:
        """
        Returns the contract given it's on chain address
        :param address: The on chain address of a contract
        :return: tuple(address, contract)
        """
        if not re.match(r"0x[a-fA-F0-9]{40}", address):
            raise CriticalError("Invalid contract address. Expected format is '0x...'.")

        if self.eth is None:
            raise CriticalError(
                "Please check whether the Infura key is set or use a different RPC method."
            )

        try:
            code = self.eth.eth_getCode(address)
        except FileNotFoundError as e:
            raise CriticalError("IPC error: " + str(e))
        except ConnectionError:
            raise CriticalError(
                "Could not connect to RPC server. Make sure that your node is running and that RPC parameters are set correctly."
            )
        except Exception as e:
            raise CriticalError("IPC / RPC error: " + str(e))

        if code == "0x" or code == "0x0":
            raise CriticalError(
                "Received an empty response from eth_getCode. Check the contract address and verify that you are on the correct chain."
            )
        else:
            self.contracts.append(
                EVMContract(
                    code, name=address, enable_online_lookup=self.enable_online_lookup
                )
            )
        return address, self.contracts[-1]  # return address and contract object

    def load_from_foundry(self):
        project_root = os.getcwd()

        cmd = ["forge", "build", "--build-info", "--force"]

        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=project_root,
            executable=shutil.which(cmd[0]),
        ) as p:

            stdout, stderr = p.communicate()
            stdout, stderr = (stdout.decode(), stderr.decode())
            if stderr:
                log.error(stderr)

            build_dir = Path(project_root, "artifacts", "contracts", "build-info")

        build_dir = os.path.join(project_root, "artifacts", "contracts", "build-info")

        files = os.listdir(build_dir)
        address = util.get_indexed_address(0)

        files = sorted(
            os.listdir(build_dir), key=lambda x: os.path.getmtime(Path(build_dir, x))
        )

        files = [str(f) for f in files if str(f).endswith(".json")]
        if not files:
            txt = f"`compile` failed. Can you run it?\n{build_dir} is empty"
            raise Exception(txt)
        contracts = []
        for file in files:
            build_info = Path(build_dir, file)

            uniq_id = file if ".json" not in file else file[0:-5]

            with open(build_info, encoding="utf8") as file_desc:
                loaded_json = json.load(file_desc)

                targets_json = loaded_json["output"]

                version_from_config = loaded_json["solcVersion"]
                input_json = loaded_json["input"]
                compiler = "solc" if input_json["language"] == "Solidity" else "vyper"
                optimizer = input_json["settings"]["optimizer"]["enabled"]

                if compiler == "vyper":
                    raise NotImplementedError("Support for Vyper is not implemented.")

                if "contracts" in targets_json:
                    for original_filename, contracts_info in targets_json[
                        "contracts"
                    ].items():
                        for contract in get_contracts_from_foundry(
                            original_filename, targets_json
                        ):
                            self.contracts.append(contract)
                            contracts.append(contract)
                            self.sigs.add_sigs(original_filename, targets_json)
        return address, contracts

    def load_from_solidity(
        self, solidity_files: List[str]
    ) -> Tuple[str, List[SolidityContract]]:
        """

        :param solidity_files: List of solidity_files
        :return: tuple of address, contract class list
        """
        address = util.get_indexed_address(0)
        contracts = []
        for file in solidity_files:
            if ":" in file:
                file, contract_name = file.split(":")
            else:
                contract_name = None

            file = os.path.expanduser(file)
            solc_binary = self.solc_binary or util.extract_binary(file)
            try:
                # import signatures from solidity source
                self.sigs.import_solidity_file(
                    file,
                    solc_binary=solc_binary,
                    solc_settings_json=self.solc_settings_json,
                )
                if contract_name is not None:
                    contract = SolidityContract(
                        input_file=file,
                        name=contract_name,
                        solc_settings_json=self.solc_settings_json,
                        solc_binary=solc_binary,
                    )
                    self.contracts.append(contract)
                    contracts.append(contract)
                else:
                    for contract in get_contracts_from_file(
                        input_file=file,
                        solc_settings_json=self.solc_settings_json,
                        solc_binary=solc_binary,
                    ):
                        self.contracts.append(contract)
                        contracts.append(contract)

            except FileNotFoundError as e:
                raise CriticalError(f"Input file not found {e}")
            except CompilerError as e:
                error_msg = str(e)
                # Check if error is related to solidity version mismatch
                if (
                    "Error: Source file requires different compiler version"
                    in error_msg
                ):
                    # Grab relevant line "pragma solidity <solv>...", excluding any comments
                    solv_pragma_line = error_msg.split("\n")[-3].split("//")[0]
                    # Grab solidity version from relevant line
                    solv_match = re.findall(r"[0-9]+\.[0-9]+\.[0-9]+", solv_pragma_line)
                    error_suggestion = (
                        "<version_number>" if len(solv_match) != 1 else solv_match[0]
                    )
                    error_msg = (
                        error_msg
                        + '\nSolidityVersionMismatch: Try adding the option "--solv '
                        + error_suggestion
                        + '"\n'
                    )

                raise CriticalError(error_msg)
            except NoContractFoundError:
                log.error(
                    "The file " + file + " does not contain a compilable contract."
                )

        return address, contracts

    @staticmethod
    def hash_for_function_signature(func: str) -> str:
        """
        Return function nadmes corresponding signature hash
        :param func: function name
        :return: Its hash signature
        """
        return "0x%s" % sha3(func)[:4].hex()

    def get_state_variable_from_storage(
        self, address: str, params: Optional[List[str]] = None
    ) -> str:
        """
        Get variables from the storage
        :param address: The contract address
        :param params: The list of parameters param types: [position, length] or ["mapping", position, key1, key2, ...  ]
                       or [position, length, array]
        :return: The corresponding storage slot and its value
        """
        params = params or []
        (position, length, mappings) = (0, 1, [])
        try:
            if params[0] == "mapping":
                if len(params) < 3:
                    raise CriticalError("Invalid number of parameters.")
                position = int(params[1])
                position_formatted = zpad(int_to_big_endian(position), 32)
                for i in range(2, len(params)):
                    key = bytes(params[i], "utf8")
                    key_formatted = rzpad(key, 32)
                    mappings.append(
                        int.from_bytes(
                            sha3(key_formatted + position_formatted), byteorder="big"
                        )
                    )

                length = len(mappings)
                if length == 1:
                    position = mappings[0]

            else:
                if len(params) >= 4:
                    raise CriticalError("Invalid number of parameters.")

                if len(params) >= 1:
                    position = int(params[0])
                if len(params) >= 2:
                    length = int(params[1])
                if len(params) == 3 and params[2] == "array":
                    position_formatted = zpad(int_to_big_endian(position), 32)
                    position = int.from_bytes(sha3(position_formatted), byteorder="big")

        except ValueError:
            raise CriticalError(
                "Invalid storage index. Please provide a numeric value."
            )

        outtxt = []

        try:
            if length == 1:
                outtxt.append(
                    "{}: {}".format(
                        position, self.eth.eth_getStorageAt(address, position)
                    )
                )
            else:
                if len(mappings) > 0:
                    for i in range(0, len(mappings)):
                        position = mappings[i]
                        outtxt.append(
                            "{}: {}".format(
                                hex(position),
                                self.eth.eth_getStorageAt(address, position),
                            )
                        )
                else:
                    for i in range(position, position + length):
                        outtxt.append(
                            "{}: {}".format(
                                hex(i), self.eth.eth_getStorageAt(address, i)
                            )
                        )
        except FileNotFoundError as e:
            raise CriticalError("IPC error: " + str(e))
        except ConnectionError:
            raise CriticalError(
                "Could not connect to RPC server. "
                "Make sure that your node is running and that RPC parameters are set correctly."
            )
        return "\n".join(outtxt)



from typing import List

from mythril.analysis.report import Issue
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.smt import Bool
from copy import deepcopy


class IssueAnnotation(StateAnnotation):
    def __init__(self, conditions: List[Bool], issue: Issue, detector):
        """
        Issue Annotation to propogate issues
        - conditions: A list of independent conditions [a, b, c, ...]
                      Each of these have to be independently be satisfied
        - issue: The issue of the annotation
        - detector: The detection module
        """
        self.conditions = conditions
        self.issue = issue
        self.detector = detector

    def persist_to_world_state(self) -> bool:
        return True

    @property
    def persist_over_calls(self) -> bool:
        return True

    def __copy__(self):
        return IssueAnnotation(
            conditions=deepcopy(self.conditions),
            issue=self.issue,
            detector=self.detector,
        )



"""This module contains the detection code for reachable exceptions."""
import logging

from typing import cast, List, Optional
from mythril.analysis import solver
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import ASSERT_VIOLATION
from mythril.exceptions import UnsatError

from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum import util
from mythril.laser.smt import And

from mythril.support.support_utils import get_code_hash

log = logging.getLogger(__name__)

# The function signature of Panic(uint256)
PANIC_SIGNATURE = [78, 72, 123, 113]


class LastJumpAnnotation(StateAnnotation):
    """State Annotation used if an overflow is both possible and used in the annotated path"""

    def __init__(self, last_jump: Optional[int] = None) -> None:
        self.last_jump: int = last_jump

    def __copy__(self):
        new_annotation = LastJumpAnnotation(self.last_jump)
        return new_annotation


class Exceptions(DetectionModule):
    """"""

    name = "Assertion violation"
    swc_id = ASSERT_VIOLATION
    description = "Checks whether any exception states are reachable."
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["INVALID", "JUMP", "REVERT"]

    def __init__(self):
        super().__init__()
        self.auto_cache = False

    def _execute(self, state: GlobalState) -> List[Issue]:
        """

        :param state:
        :return:
        """
        issues = self._analyze_state(state)
        for issue in issues:
            self.cache.add((issue.source_location, issue.bytecode_hash))
        return issues

    def _analyze_state(self, state) -> List[Issue]:
        """

        :param state:
        :return:
        """
        opcode = state.get_current_instruction()["opcode"]
        address = state.get_current_instruction()["address"]

        annotations = cast(
            List[LastJumpAnnotation],
            [a for a in state.get_annotations(LastJumpAnnotation)],
        )
        if len(annotations) == 0:
            state.annotate(LastJumpAnnotation())
            annotations = cast(
                List[LastJumpAnnotation],
                [a for a in state.get_annotations(LastJumpAnnotation)],
            )

        if opcode == "JUMP":
            annotations[0].last_jump = address
            return []

        if opcode == "REVERT" and not is_assertion_failure(state):
            return []

        cache_address = annotations[0].last_jump
        if (
            cache_address,
            get_code_hash(state.environment.code.bytecode),
        ) in self.cache:
            return []

        log.debug(
            "ASSERT_FAIL/REVERT in function " + state.environment.active_function_name
        )

        try:
            description_tail = (
                "It is possible to trigger an assertion violation. Note that Solidity assert() statements should "
                "only be used to check invariants. Review the transaction trace generated for this issue and "
                "either make sure your program logic is correct, or use require() instead of assert() if your goal "
                "is to constrain user inputs or enforce preconditions. Remember to validate inputs from both callers "
                "(for instance, via passed arguments) and callees (for instance, via return values)."
            )
            transaction_sequence = solver.get_transaction_sequence(
                state, state.world_state.constraints
            )

            issue = Issue(
                contract=state.environment.active_account.contract_name,
                function_name=state.environment.active_function_name,
                address=address,
                swc_id=ASSERT_VIOLATION,
                title="Exception State",
                severity="Medium",
                description_head="An assertion violation was triggered.",
                description_tail=description_tail,
                bytecode=state.environment.code.bytecode,
                transaction_sequence=transaction_sequence,
                gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
                source_location=cache_address,
            )
            state.annotate(
                IssueAnnotation(
                    conditions=[And(*state.world_state.constraints)],
                    issue=issue,
                    detector=self,
                )
            )

            return [issue]

        except UnsatError:
            log.debug("no model found")

        return []


def is_assertion_failure(global_state):
    state = global_state.mstate
    offset, length = state.stack[-1], state.stack[-2]
    try:
        return_data = state.memory[
            util.get_concrete_int(offset) : util.get_concrete_int(offset + length)
        ]
    except TypeError:
        return False

    return return_data[:4] == PANIC_SIGNATURE and return_data[-1] == 1


detector = Exceptions()
